from chardet.enums import CharacterCategory, SequenceLikelihood
from chardet.sbcharsetprober import SingleByteCharSetModel

GERMAN_LANG_MODEL = {
    25: {  # 'A'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.LIKELY,  # 'B'
        45: SequenceLikelihood.LIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.LIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.LIKELY,  # 'K'
        41: SequenceLikelihood.LIKELY,  # 'L'
        29: SequenceLikelihood.LIKELY,  # 'M'
        43: SequenceLikelihood.LIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.LIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.LIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.LIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.LIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    28: {  # 'B'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.LIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.LIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    45: {  # 'C'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.LIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.LIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.LIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.LIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.LIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.LIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    26: {  # 'D'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.LIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.LIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.LIKELY,  # 'U'
        42: SequenceLikelihood.LIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    30: {  # 'E'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.LIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.LIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.LIKELY,  # 'L'
        29: SequenceLikelihood.LIKELY,  # 'M'
        43: SequenceLikelihood.LIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.LIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.LIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.LIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.POSITIVE,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.NEGATIVE,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.UNLIKELY,  # 'ẞ'
    },
    33: {  # 'F'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.LIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.POSITIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    35: {  # 'G'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.LIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.LIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    38: {  # 'H'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    37: {  # 'I'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.LIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.LIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.LIKELY,  # 'L'
        29: SequenceLikelihood.LIKELY,  # 'M'
        43: SequenceLikelihood.LIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.LIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.LIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.LIKELY,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.LIKELY,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.UNLIKELY,  # 'ẞ'
    },
    51: {  # 'J'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    32: {  # 'K'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.LIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    41: {  # 'L'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.LIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    29: {  # 'M'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.LIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.LIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    43: {  # 'N'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.LIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    49: {  # 'O'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.LIKELY,  # 'L'
        29: SequenceLikelihood.LIKELY,  # 'M'
        43: SequenceLikelihood.LIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.LIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.LIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.LIKELY,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.NEGATIVE,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    34: {  # 'P'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.LIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.LIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    55: {  # 'Q'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.NEGATIVE,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    40: {  # 'R'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.LIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.LIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    22: {  # 'S'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.LIKELY,  # 'B'
        45: SequenceLikelihood.LIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.LIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.LIKELY,  # 'L'
        29: SequenceLikelihood.LIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.LIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.LIKELY,  # 'U'
        42: SequenceLikelihood.LIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.LIKELY,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    36: {  # 'T'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.LIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.LIKELY,  # 'U'
        42: SequenceLikelihood.LIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    46: {  # 'U'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.LIKELY,  # 'M'
        43: SequenceLikelihood.LIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.LIKELY,  # 'R'
        22: SequenceLikelihood.LIKELY,  # 'S'
        36: SequenceLikelihood.LIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.LIKELY,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.NEGATIVE,  # 'ä'
        39: SequenceLikelihood.NEGATIVE,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.UNLIKELY,  # 'ẞ'
    },
    42: {  # 'V'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.LIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.LIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.LIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    31: {  # 'W'
        25: SequenceLikelihood.LIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.LIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.LIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    57: {  # 'X'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.NEGATIVE,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    56: {  # 'Y'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.LIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.LIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.NEGATIVE,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    47: {  # 'Z'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.LIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.LIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    7: {  # 'a'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.LIKELY,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.LIKELY,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.POSITIVE,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.POSITIVE,  # 'x'
        44: SequenceLikelihood.POSITIVE,  # 'y'
        20: SequenceLikelihood.POSITIVE,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.POSITIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    16: {  # 'b'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.LIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    13: {  # 'c'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.LIKELY,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    8: {  # 'd'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    1: {  # 'e'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.LIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.LIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.POSITIVE,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.POSITIVE,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.POSITIVE,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.POSITIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    17: {  # 'f'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.POSITIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    14: {  # 'g'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    9: {  # 'h'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.POSITIVE,  # 'ö'
        24: SequenceLikelihood.POSITIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    3: {  # 'i'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.LIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.LIKELY,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.LIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.POSITIVE,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.LIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.POSITIVE,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.POSITIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    50: {  # 'j'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.LIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.LIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    18: {  # 'k'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.POSITIVE,  # 'ö'
        24: SequenceLikelihood.POSITIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    11: {  # 'l'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.POSITIVE,  # 'y'
        20: SequenceLikelihood.POSITIVE,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.POSITIVE,  # 'ö'
        24: SequenceLikelihood.POSITIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    15: {  # 'm'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.POSITIVE,  # 'ö'
        24: SequenceLikelihood.POSITIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    2: {  # 'n'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.POSITIVE,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.POSITIVE,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    12: {  # 'o'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.LIKELY,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.POSITIVE,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.LIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.POSITIVE,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.POSITIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    21: {  # 'p'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    53: {  # 'q'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.NEGATIVE,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    4: {  # 'r'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.LIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.LIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.POSITIVE,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.POSITIVE,  # 'y'
        20: SequenceLikelihood.POSITIVE,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.POSITIVE,  # 'ö'
        24: SequenceLikelihood.POSITIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    6: {  # 's'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.LIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.POSITIVE,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.POSITIVE,  # 'y'
        20: SequenceLikelihood.POSITIVE,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    5: {  # 't'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.LIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.POSITIVE,  # 'y'
        20: SequenceLikelihood.POSITIVE,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.POSITIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    10: {  # 'u'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.POSITIVE,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.POSITIVE,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.LIKELY,  # 'o'
        21: SequenceLikelihood.POSITIVE,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.LIKELY,  # 'w'
        52: SequenceLikelihood.LIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.POSITIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    23: {  # 'v'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    19: {  # 'w'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.POSITIVE,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.POSITIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    52: {  # 'x'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.LIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.LIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.LIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    44: {  # 'y'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.LIKELY,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.LIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.LIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.LIKELY,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.LIKELY,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.LIKELY,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    20: {  # 'z'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.UNLIKELY,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.POSITIVE,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.LIKELY,  # 'h'
        3: SequenceLikelihood.POSITIVE,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.POSITIVE,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.LIKELY,  # 'v'
        19: SequenceLikelihood.POSITIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.LIKELY,  # 'y'
        20: SequenceLikelihood.LIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.LIKELY,  # 'ä'
        39: SequenceLikelihood.LIKELY,  # 'ö'
        24: SequenceLikelihood.LIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    59: {  # 'Ä'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.NEGATIVE,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.NEGATIVE,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.NEGATIVE,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.LIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.NEGATIVE,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.NEGATIVE,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    58: {  # 'Ö'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.UNLIKELY,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.NEGATIVE,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.NEGATIVE,  # 'X'
        56: SequenceLikelihood.NEGATIVE,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.UNLIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.NEGATIVE,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.LIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.NEGATIVE,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.NEGATIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.NEGATIVE,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.NEGATIVE,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.UNLIKELY,  # 'Ä'
        58: SequenceLikelihood.UNLIKELY,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.NEGATIVE,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    54: {  # 'Ü'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.NEGATIVE,  # 'I'
        51: SequenceLikelihood.NEGATIVE,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.NEGATIVE,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.NEGATIVE,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.NEGATIVE,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.UNLIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.NEGATIVE,  # 'j'
        18: SequenceLikelihood.NEGATIVE,  # 'k'
        11: SequenceLikelihood.UNLIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.UNLIKELY,  # 'n'
        12: SequenceLikelihood.NEGATIVE,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.NEGATIVE,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.UNLIKELY,  # 's'
        5: SequenceLikelihood.UNLIKELY,  # 't'
        10: SequenceLikelihood.NEGATIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.NEGATIVE,  # 'w'
        52: SequenceLikelihood.NEGATIVE,  # 'x'
        44: SequenceLikelihood.NEGATIVE,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.NEGATIVE,  # 'ä'
        39: SequenceLikelihood.NEGATIVE,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    48: {  # 'ß'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.NEGATIVE,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.NEGATIVE,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.UNLIKELY,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.NEGATIVE,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.NEGATIVE,  # 'X'
        56: SequenceLikelihood.NEGATIVE,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.LIKELY,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.UNLIKELY,  # 'c'
        8: SequenceLikelihood.UNLIKELY,  # 'd'
        1: SequenceLikelihood.POSITIVE,  # 'e'
        17: SequenceLikelihood.UNLIKELY,  # 'f'
        14: SequenceLikelihood.LIKELY,  # 'g'
        9: SequenceLikelihood.UNLIKELY,  # 'h'
        3: SequenceLikelihood.LIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.LIKELY,  # 'l'
        15: SequenceLikelihood.UNLIKELY,  # 'm'
        2: SequenceLikelihood.LIKELY,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.UNLIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.UNLIKELY,  # 'r'
        6: SequenceLikelihood.LIKELY,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.UNLIKELY,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    27: {  # 'ä'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.NEGATIVE,  # 'I'
        51: SequenceLikelihood.NEGATIVE,  # 'J'
        32: SequenceLikelihood.NEGATIVE,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.NEGATIVE,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.NEGATIVE,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.NEGATIVE,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.NEGATIVE,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.POSITIVE,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.LIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.POSITIVE,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.LIKELY,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.UNLIKELY,  # 'ẞ'
    },
    39: {  # 'ö'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.NEGATIVE,  # 'E'
        33: SequenceLikelihood.UNLIKELY,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.NEGATIVE,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.NEGATIVE,  # 'O'
        34: SequenceLikelihood.UNLIKELY,  # 'P'
        55: SequenceLikelihood.NEGATIVE,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.NEGATIVE,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.NEGATIVE,  # 'W'
        57: SequenceLikelihood.NEGATIVE,  # 'X'
        56: SequenceLikelihood.NEGATIVE,  # 'Y'
        47: SequenceLikelihood.NEGATIVE,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.LIKELY,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.POSITIVE,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.LIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.UNLIKELY,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.LIKELY,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.POSITIVE,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    24: {  # 'ü'
        25: SequenceLikelihood.UNLIKELY,  # 'A'
        28: SequenceLikelihood.UNLIKELY,  # 'B'
        45: SequenceLikelihood.UNLIKELY,  # 'C'
        26: SequenceLikelihood.UNLIKELY,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.NEGATIVE,  # 'F'
        35: SequenceLikelihood.UNLIKELY,  # 'G'
        38: SequenceLikelihood.UNLIKELY,  # 'H'
        37: SequenceLikelihood.UNLIKELY,  # 'I'
        51: SequenceLikelihood.UNLIKELY,  # 'J'
        32: SequenceLikelihood.UNLIKELY,  # 'K'
        41: SequenceLikelihood.UNLIKELY,  # 'L'
        29: SequenceLikelihood.UNLIKELY,  # 'M'
        43: SequenceLikelihood.UNLIKELY,  # 'N'
        49: SequenceLikelihood.NEGATIVE,  # 'O'
        34: SequenceLikelihood.NEGATIVE,  # 'P'
        55: SequenceLikelihood.NEGATIVE,  # 'Q'
        40: SequenceLikelihood.UNLIKELY,  # 'R'
        22: SequenceLikelihood.UNLIKELY,  # 'S'
        36: SequenceLikelihood.UNLIKELY,  # 'T'
        46: SequenceLikelihood.UNLIKELY,  # 'U'
        42: SequenceLikelihood.UNLIKELY,  # 'V'
        31: SequenceLikelihood.UNLIKELY,  # 'W'
        57: SequenceLikelihood.UNLIKELY,  # 'X'
        56: SequenceLikelihood.NEGATIVE,  # 'Y'
        47: SequenceLikelihood.UNLIKELY,  # 'Z'
        7: SequenceLikelihood.UNLIKELY,  # 'a'
        16: SequenceLikelihood.POSITIVE,  # 'b'
        13: SequenceLikelihood.POSITIVE,  # 'c'
        8: SequenceLikelihood.LIKELY,  # 'd'
        1: SequenceLikelihood.UNLIKELY,  # 'e'
        17: SequenceLikelihood.LIKELY,  # 'f'
        14: SequenceLikelihood.POSITIVE,  # 'g'
        9: SequenceLikelihood.POSITIVE,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.UNLIKELY,  # 'j'
        18: SequenceLikelihood.UNLIKELY,  # 'k'
        11: SequenceLikelihood.POSITIVE,  # 'l'
        15: SequenceLikelihood.LIKELY,  # 'm'
        2: SequenceLikelihood.POSITIVE,  # 'n'
        12: SequenceLikelihood.UNLIKELY,  # 'o'
        21: SequenceLikelihood.LIKELY,  # 'p'
        53: SequenceLikelihood.NEGATIVE,  # 'q'
        4: SequenceLikelihood.POSITIVE,  # 'r'
        6: SequenceLikelihood.POSITIVE,  # 's'
        5: SequenceLikelihood.POSITIVE,  # 't'
        10: SequenceLikelihood.UNLIKELY,  # 'u'
        23: SequenceLikelihood.UNLIKELY,  # 'v'
        19: SequenceLikelihood.UNLIKELY,  # 'w'
        52: SequenceLikelihood.UNLIKELY,  # 'x'
        44: SequenceLikelihood.UNLIKELY,  # 'y'
        20: SequenceLikelihood.UNLIKELY,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.UNLIKELY,  # 'Ü'
        48: SequenceLikelihood.LIKELY,  # 'ß'
        27: SequenceLikelihood.UNLIKELY,  # 'ä'
        39: SequenceLikelihood.UNLIKELY,  # 'ö'
        24: SequenceLikelihood.UNLIKELY,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
    60: {  # 'ẞ'
        25: SequenceLikelihood.NEGATIVE,  # 'A'
        28: SequenceLikelihood.NEGATIVE,  # 'B'
        45: SequenceLikelihood.NEGATIVE,  # 'C'
        26: SequenceLikelihood.NEGATIVE,  # 'D'
        30: SequenceLikelihood.UNLIKELY,  # 'E'
        33: SequenceLikelihood.NEGATIVE,  # 'F'
        35: SequenceLikelihood.NEGATIVE,  # 'G'
        38: SequenceLikelihood.NEGATIVE,  # 'H'
        37: SequenceLikelihood.NEGATIVE,  # 'I'
        51: SequenceLikelihood.NEGATIVE,  # 'J'
        32: SequenceLikelihood.NEGATIVE,  # 'K'
        41: SequenceLikelihood.NEGATIVE,  # 'L'
        29: SequenceLikelihood.NEGATIVE,  # 'M'
        43: SequenceLikelihood.NEGATIVE,  # 'N'
        49: SequenceLikelihood.NEGATIVE,  # 'O'
        34: SequenceLikelihood.NEGATIVE,  # 'P'
        55: SequenceLikelihood.NEGATIVE,  # 'Q'
        40: SequenceLikelihood.NEGATIVE,  # 'R'
        22: SequenceLikelihood.NEGATIVE,  # 'S'
        36: SequenceLikelihood.NEGATIVE,  # 'T'
        46: SequenceLikelihood.NEGATIVE,  # 'U'
        42: SequenceLikelihood.NEGATIVE,  # 'V'
        31: SequenceLikelihood.NEGATIVE,  # 'W'
        57: SequenceLikelihood.NEGATIVE,  # 'X'
        56: SequenceLikelihood.NEGATIVE,  # 'Y'
        47: SequenceLikelihood.NEGATIVE,  # 'Z'
        7: SequenceLikelihood.NEGATIVE,  # 'a'
        16: SequenceLikelihood.NEGATIVE,  # 'b'
        13: SequenceLikelihood.NEGATIVE,  # 'c'
        8: SequenceLikelihood.NEGATIVE,  # 'd'
        1: SequenceLikelihood.NEGATIVE,  # 'e'
        17: SequenceLikelihood.NEGATIVE,  # 'f'
        14: SequenceLikelihood.NEGATIVE,  # 'g'
        9: SequenceLikelihood.NEGATIVE,  # 'h'
        3: SequenceLikelihood.UNLIKELY,  # 'i'
        50: SequenceLikelihood.NEGATIVE,  # 'j'
        18: SequenceLikelihood.NEGATIVE,  # 'k'
        11: SequenceLikelihood.NEGATIVE,  # 'l'
        15: SequenceLikelihood.NEGATIVE,  # 'm'
        2: SequenceLikelihood.NEGATIVE,  # 'n'
        12: SequenceLikelihood.NEGATIVE,  # 'o'
        21: SequenceLikelihood.NEGATIVE,  # 'p'
        53: SequenceLikelihood.NEGATIVE,  # 'q'
        4: SequenceLikelihood.NEGATIVE,  # 'r'
        6: SequenceLikelihood.NEGATIVE,  # 's'
        5: SequenceLikelihood.NEGATIVE,  # 't'
        10: SequenceLikelihood.NEGATIVE,  # 'u'
        23: SequenceLikelihood.NEGATIVE,  # 'v'
        19: SequenceLikelihood.NEGATIVE,  # 'w'
        52: SequenceLikelihood.NEGATIVE,  # 'x'
        44: SequenceLikelihood.NEGATIVE,  # 'y'
        20: SequenceLikelihood.NEGATIVE,  # 'z'
        59: SequenceLikelihood.NEGATIVE,  # 'Ä'
        58: SequenceLikelihood.NEGATIVE,  # 'Ö'
        54: SequenceLikelihood.NEGATIVE,  # 'Ü'
        48: SequenceLikelihood.NEGATIVE,  # 'ß'
        27: SequenceLikelihood.NEGATIVE,  # 'ä'
        39: SequenceLikelihood.NEGATIVE,  # 'ö'
        24: SequenceLikelihood.NEGATIVE,  # 'ü'
        60: SequenceLikelihood.NEGATIVE,  # 'ẞ'
    },
}

# Character Mapping Table(s):
CP037_GERMAN_CHAR_TO_ORDER = {
    0: CharacterCategory.CONTROL,  # '\x00'
    1: CharacterCategory.CONTROL,  # '\x01'
    2: CharacterCategory.CONTROL,  # '\x02'
    3: CharacterCategory.CONTROL,  # '\x03'
    4: CharacterCategory.CONTROL,  # '\x9c'
    5: CharacterCategory.CONTROL,  # '\t'
    6: CharacterCategory.CONTROL,  # '\x86'
    7: CharacterCategory.CONTROL,  # '\x7f'
    8: CharacterCategory.CONTROL,  # '\x97'
    9: CharacterCategory.CONTROL,  # '\x8d'
    10: CharacterCategory.CONTROL,  # '\x8e'
    11: CharacterCategory.CONTROL,  # '\x0b'
    12: CharacterCategory.CONTROL,  # '\x0c'
    13: CharacterCategory.LINE_BREAK,  # '\r'
    14: CharacterCategory.CONTROL,  # '\x0e'
    15: CharacterCategory.CONTROL,  # '\x0f'
    16: CharacterCategory.CONTROL,  # '\x10'
    17: CharacterCategory.CONTROL,  # '\x11'
    18: CharacterCategory.CONTROL,  # '\x12'
    19: CharacterCategory.CONTROL,  # '\x13'
    20: CharacterCategory.CONTROL,  # '\x9d'
    21: CharacterCategory.CONTROL,  # '\x85'
    22: CharacterCategory.CONTROL,  # '\x08'
    23: CharacterCategory.CONTROL,  # '\x87'
    24: CharacterCategory.CONTROL,  # '\x18'
    25: CharacterCategory.CONTROL,  # '\x19'
    26: CharacterCategory.CONTROL,  # '\x92'
    27: CharacterCategory.CONTROL,  # '\x8f'
    28: CharacterCategory.CONTROL,  # '\x1c'
    29: CharacterCategory.CONTROL,  # '\x1d'
    30: CharacterCategory.CONTROL,  # '\x1e'
    31: CharacterCategory.CONTROL,  # '\x1f'
    32: CharacterCategory.CONTROL,  # '\x80'
    33: CharacterCategory.CONTROL,  # '\x81'
    34: CharacterCategory.CONTROL,  # '\x82'
    35: CharacterCategory.CONTROL,  # '\x83'
    36: CharacterCategory.CONTROL,  # '\x84'
    37: CharacterCategory.LINE_BREAK,  # '\n'
    38: CharacterCategory.CONTROL,  # '\x17'
    39: CharacterCategory.CONTROL,  # '\x1b'
    40: CharacterCategory.CONTROL,  # '\x88'
    41: CharacterCategory.CONTROL,  # '\x89'
    42: CharacterCategory.CONTROL,  # '\x8a'
    43: CharacterCategory.CONTROL,  # '\x8b'
    44: CharacterCategory.CONTROL,  # '\x8c'
    45: CharacterCategory.CONTROL,  # '\x05'
    46: CharacterCategory.CONTROL,  # '\x06'
    47: CharacterCategory.CONTROL,  # '\x07'
    48: CharacterCategory.CONTROL,  # '\x90'
    49: CharacterCategory.CONTROL,  # '\x91'
    50: CharacterCategory.CONTROL,  # '\x16'
    51: CharacterCategory.CONTROL,  # '\x93'
    52: CharacterCategory.CONTROL,  # '\x94'
    53: CharacterCategory.CONTROL,  # '\x95'
    54: CharacterCategory.CONTROL,  # '\x96'
    55: CharacterCategory.CONTROL,  # '\x04'
    56: CharacterCategory.CONTROL,  # '\x98'
    57: CharacterCategory.CONTROL,  # '\x99'
    58: CharacterCategory.CONTROL,  # '\x9a'
    59: CharacterCategory.CONTROL,  # '\x9b'
    60: CharacterCategory.CONTROL,  # '\x14'
    61: CharacterCategory.CONTROL,  # '\x15'
    62: CharacterCategory.CONTROL,  # '\x9e'
    63: CharacterCategory.CONTROL,  # '\x1a'
    64: CharacterCategory.CONTROL,  # ' '
    65: CharacterCategory.CONTROL,  # '\xa0'
    66: 0,  # 'â'
    67: 27,  # 'ä'
    68: 0,  # 'à'
    69: 0,  # 'á'
    70: 0,  # 'ã'
    71: 0,  # 'å'
    72: 0,  # 'ç'
    73: 0,  # 'ñ'
    74: CharacterCategory.SYMBOL,  # '¢'
    75: CharacterCategory.SYMBOL,  # '.'
    76: CharacterCategory.SYMBOL,  # '<'
    77: CharacterCategory.SYMBOL,  # '('
    78: CharacterCategory.SYMBOL,  # '+'
    79: CharacterCategory.SYMBOL,  # '|'
    80: CharacterCategory.SYMBOL,  # '&'
    81: 0,  # 'é'
    82: 0,  # 'ê'
    83: 0,  # 'ë'
    84: 0,  # 'è'
    85: 0,  # 'í'
    86: 0,  # 'î'
    87: 0,  # 'ï'
    88: 0,  # 'ì'
    89: 48,  # 'ß'
    90: CharacterCategory.SYMBOL,  # '!'
    91: CharacterCategory.SYMBOL,  # '$'
    92: CharacterCategory.SYMBOL,  # '*'
    93: CharacterCategory.SYMBOL,  # ')'
    94: CharacterCategory.SYMBOL,  # ';'
    95: CharacterCategory.SYMBOL,  # '¬'
    96: CharacterCategory.SYMBOL,  # '-'
    97: CharacterCategory.SYMBOL,  # '/'
    98: 0,  # 'Â'
    99: 59,  # 'Ä'
    100: 0,  # 'À'
    101: 0,  # 'Á'
    102: 0,  # 'Ã'
    103: 0,  # 'Å'
    104: 0,  # 'Ç'
    105: 0,  # 'Ñ'
    106: CharacterCategory.SYMBOL,  # '¦'
    107: CharacterCategory.SYMBOL,  # ','
    108: CharacterCategory.SYMBOL,  # '%'
    109: CharacterCategory.SYMBOL,  # '_'
    110: CharacterCategory.SYMBOL,  # '>'
    111: CharacterCategory.SYMBOL,  # '?'
    112: 0,  # 'ø'
    113: 0,  # 'É'
    114: 0,  # 'Ê'
    115: 0,  # 'Ë'
    116: 0,  # 'È'
    117: 0,  # 'Í'
    118: 0,  # 'Î'
    119: 0,  # 'Ï'
    120: 0,  # 'Ì'
    121: CharacterCategory.SYMBOL,  # '`'
    122: CharacterCategory.SYMBOL,  # ':'
    123: CharacterCategory.SYMBOL,  # '#'
    124: CharacterCategory.SYMBOL,  # '@'
    125: CharacterCategory.SYMBOL,  # "'"
    126: CharacterCategory.SYMBOL,  # '='
    127: CharacterCategory.SYMBOL,  # '"'
    128: 0,  # 'Ø'
    129: 7,  # 'a'
    130: 16,  # 'b'
    131: 13,  # 'c'
    132: 8,  # 'd'
    133: 1,  # 'e'
    134: 17,  # 'f'
    135: 14,  # 'g'
    136: 9,  # 'h'
    137: 3,  # 'i'
    138: CharacterCategory.SYMBOL,  # '«'
    139: CharacterCategory.SYMBOL,  # '»'
    140: 0,  # 'ð'
    141: 0,  # 'ý'
    142: 0,  # 'þ'
    143: CharacterCategory.SYMBOL,  # '±'
    144: CharacterCategory.SYMBOL,  # '°'
    145: 50,  # 'j'
    146: 18,  # 'k'
    147: 11,  # 'l'
    148: 15,  # 'm'
    149: 2,  # 'n'
    150: 12,  # 'o'
    151: 21,  # 'p'
    152: 53,  # 'q'
    153: 4,  # 'r'
    154: 0,  # 'ª'
    155: 0,  # 'º'
    156: 0,  # 'æ'
    157: CharacterCategory.SYMBOL,  # '¸'
    158: 0,  # 'Æ'
    159: CharacterCategory.SYMBOL,  # '¤'
    160: 0,  # 'µ'
    161: CharacterCategory.SYMBOL,  # '~'
    162: 6,  # 's'
    163: 5,  # 't'
    164: 10,  # 'u'
    165: 23,  # 'v'
    166: 19,  # 'w'
    167: 52,  # 'x'
    168: 44,  # 'y'
    169: 20,  # 'z'
    170: CharacterCategory.SYMBOL,  # '¡'
    171: CharacterCategory.SYMBOL,  # '¿'
    172: 0,  # 'Ð'
    173: 0,  # 'Ý'
    174: 0,  # 'Þ'
    175: CharacterCategory.SYMBOL,  # '®'
    176: CharacterCategory.SYMBOL,  # '^'
    177: CharacterCategory.SYMBOL,  # '£'
    178: CharacterCategory.SYMBOL,  # '¥'
    179: CharacterCategory.SYMBOL,  # '·'
    180: CharacterCategory.SYMBOL,  # '©'
    181: CharacterCategory.SYMBOL,  # '§'
    182: CharacterCategory.SYMBOL,  # '¶'
    183: CharacterCategory.DIGIT,  # '¼'
    184: CharacterCategory.DIGIT,  # '½'
    185: CharacterCategory.DIGIT,  # '¾'
    186: CharacterCategory.SYMBOL,  # '['
    187: CharacterCategory.SYMBOL,  # ']'
    188: CharacterCategory.SYMBOL,  # '¯'
    189: CharacterCategory.SYMBOL,  # '¨'
    190: CharacterCategory.SYMBOL,  # '´'
    191: CharacterCategory.SYMBOL,  # '×'
    192: CharacterCategory.SYMBOL,  # '{'
    193: 25,  # 'A'
    194: 28,  # 'B'
    195: 45,  # 'C'
    196: 26,  # 'D'
    197: 30,  # 'E'
    198: 33,  # 'F'
    199: 35,  # 'G'
    200: 38,  # 'H'
    201: 37,  # 'I'
    202: CharacterCategory.CONTROL,  # '\xad'
    203: 0,  # 'ô'
    204: 39,  # 'ö'
    205: 0,  # 'ò'
    206: 0,  # 'ó'
    207: 0,  # 'õ'
    208: CharacterCategory.SYMBOL,  # '}'
    209: 51,  # 'J'
    210: 32,  # 'K'
    211: 41,  # 'L'
    212: 29,  # 'M'
    213: 43,  # 'N'
    214: 49,  # 'O'
    215: 34,  # 'P'
    216: 55,  # 'Q'
    217: 40,  # 'R'
    218: CharacterCategory.DIGIT,  # '¹'
    219: 0,  # 'û'
    220: 24,  # 'ü'
    221: 0,  # 'ù'
    222: 0,  # 'ú'
    223: 0,  # 'ÿ'
    224: CharacterCategory.SYMBOL,  # '\\'
    225: CharacterCategory.SYMBOL,  # '÷'
    226: 22,  # 'S'
    227: 36,  # 'T'
    228: 46,  # 'U'
    229: 42,  # 'V'
    230: 31,  # 'W'
    231: 57,  # 'X'
    232: 56,  # 'Y'
    233: 47,  # 'Z'
    234: CharacterCategory.DIGIT,  # '²'
    235: 0,  # 'Ô'
    236: 58,  # 'Ö'
    237: 0,  # 'Ò'
    238: 0,  # 'Ó'
    239: 0,  # 'Õ'
    240: CharacterCategory.DIGIT,  # '0'
    241: CharacterCategory.DIGIT,  # '1'
    242: CharacterCategory.DIGIT,  # '2'
    243: CharacterCategory.DIGIT,  # '3'
    244: CharacterCategory.DIGIT,  # '4'
    245: CharacterCategory.DIGIT,  # '5'
    246: CharacterCategory.DIGIT,  # '6'
    247: CharacterCategory.DIGIT,  # '7'
    248: CharacterCategory.DIGIT,  # '8'
    249: CharacterCategory.DIGIT,  # '9'
    250: CharacterCategory.DIGIT,  # '³'
    251: 0,  # 'Û'
    252: 54,  # 'Ü'
    253: 0,  # 'Ù'
    254: 0,  # 'Ú'
    255: CharacterCategory.CONTROL,  # '\x9f'
}

CP037_GERMAN_MODEL = SingleByteCharSetModel(
    charset_name="CP037",
    language="German",
    char_to_order_map=CP037_GERMAN_CHAR_TO_ORDER,
    language_model=GERMAN_LANG_MODEL,
    typical_positive_ratio=0.9706854667596968,
    keep_ascii_letters=True,
    alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyzÄÖÜßäöüẞ",
)

CP437_GERMAN_CHAR_TO_ORDER = {
    0: CharacterCategory.CONTROL,  # '\x00'
    1: CharacterCategory.CONTROL,  # '\x01'
    2: CharacterCategory.CONTROL,  # '\x02'
    3: CharacterCategory.CONTROL,  # '\x03'
    4: CharacterCategory.CONTROL,  # '\x04'
    5: CharacterCategory.CONTROL,  # '\x05'
    6: CharacterCategory.CONTROL,  # '\x06'
    7: CharacterCategory.CONTROL,  # '\x07'
    8: CharacterCategory.CONTROL,  # '\x08'
    9: CharacterCategory.CONTROL,  # '\t'
    10: CharacterCategory.LINE_BREAK,  # '\n'
    11: CharacterCategory.CONTROL,  # '\x0b'
    12: CharacterCategory.CONTROL,  # '\x0c'
    13: CharacterCategory.LINE_BREAK,  # '\r'
    14: CharacterCategory.CONTROL,  # '\x0e'
    15: CharacterCategory.CONTROL,  # '\x0f'
    16: CharacterCategory.CONTROL,  # '\x10'
    17: CharacterCategory.CONTROL,  # '\x11'
    18: CharacterCategory.CONTROL,  # '\x12'
    19: CharacterCategory.CONTROL,  # '\x13'
    20: CharacterCategory.CONTROL,  # '\x14'
    21: CharacterCategory.CONTROL,  # '\x15'
    22: CharacterCategory.CONTROL,  # '\x16'
    23: CharacterCategory.CONTROL,  # '\x17'
    24: CharacterCategory.CONTROL,  # '\x18'
    25: CharacterCategory.CONTROL,  # '\x19'
    26: CharacterCategory.CONTROL,  # '\x1a'
    27: CharacterCategory.CONTROL,  # '\x1b'
    28: CharacterCategory.CONTROL,  # '\x1c'
    29: CharacterCategory.CONTROL,  # '\x1d'
    30: CharacterCategory.CONTROL,  # '\x1e'
    31: CharacterCategory.CONTROL,  # '\x1f'
    32: CharacterCategory.CONTROL,  # ' '
    33: CharacterCategory.SYMBOL,  # '!'
    34: CharacterCategory.SYMBOL,  # '"'
    35: CharacterCategory.SYMBOL,  # '#'
    36: CharacterCategory.SYMBOL,  # '$'
    37: CharacterCategory.SYMBOL,  # '%'
    38: CharacterCategory.SYMBOL,  # '&'
    39: CharacterCategory.SYMBOL,  # "'"
    40: CharacterCategory.SYMBOL,  # '('
    41: CharacterCategory.SYMBOL,  # ')'
    42: CharacterCategory.SYMBOL,  # '*'
    43: CharacterCategory.SYMBOL,  # '+'
    44: CharacterCategory.SYMBOL,  # ','
    45: CharacterCategory.SYMBOL,  # '-'
    46: CharacterCategory.SYMBOL,  # '.'
    47: CharacterCategory.SYMBOL,  # '/'
    48: CharacterCategory.DIGIT,  # '0'
    49: CharacterCategory.DIGIT,  # '1'
    50: CharacterCategory.DIGIT,  # '2'
    51: CharacterCategory.DIGIT,  # '3'
    52: CharacterCategory.DIGIT,  # '4'
    53: CharacterCategory.DIGIT,  # '5'
    54: CharacterCategory.DIGIT,  # '6'
    55: CharacterCategory.DIGIT,  # '7'
    56: CharacterCategory.DIGIT,  # '8'
    57: CharacterCategory.DIGIT,  # '9'
    58: CharacterCategory.SYMBOL,  # ':'
    59: CharacterCategory.SYMBOL,  # ';'
    60: CharacterCategory.SYMBOL,  # '<'
    61: CharacterCategory.SYMBOL,  # '='
    62: CharacterCategory.SYMBOL,  # '>'
    63: CharacterCategory.SYMBOL,  # '?'
    64: CharacterCategory.SYMBOL,  # '@'
    65: 25,  # 'A'
    66: 28,  # 'B'
    67: 45,  # 'C'
    68: 26,  # 'D'
    69: 30,  # 'E'
    70: 33,  # 'F'
    71: 35,  # 'G'
    72: 38,  # 'H'
    73: 37,  # 'I'
    74: 51,  # 'J'
    75: 32,  # 'K'
    76: 41,  # 'L'
    77: 29,  # 'M'
    78: 43,  # 'N'
    79: 49,  # 'O'
    80: 34,  # 'P'
    81: 55,  # 'Q'
    82: 40,  # 'R'
    83: 22,  # 'S'
    84: 36,  # 'T'
    85: 46,  # 'U'
    86: 42,  # 'V'
    87: 31,  # 'W'
    88: 57,  # 'X'
    89: 56,  # 'Y'
    90: 47,  # 'Z'
    91: CharacterCategory.SYMBOL,  # '['
    92: CharacterCategory.SYMBOL,  # '\\'
    93: CharacterCategory.SYMBOL,  # ']'
    94: CharacterCategory.SYMBOL,  # '^'
    95: CharacterCategory.SYMBOL,  # '_'
    96: CharacterCategory.SYMBOL,  # '`'
    97: 7,  # 'a'
    98: 16,  # 'b'
    99: 13,  # 'c'
    100: 8,  # 'd'
    101: 1,  # 'e'
    102: 17,  # 'f'
    103: 14,  # 'g'
    104: 9,  # 'h'
    105: 3,  # 'i'
    106: 50,  # 'j'
    107: 18,  # 'k'
    108: 11,  # 'l'
    109: 15,  # 'm'
    110: 2,  # 'n'
    111: 12,  # 'o'
    112: 21,  # 'p'
    113: 53,  # 'q'
    114: 4,  # 'r'
    115: 6,  # 's'
    116: 5,  # 't'
    117: 10,  # 'u'
    118: 23,  # 'v'
    119: 19,  # 'w'
    120: 52,  # 'x'
    121: 44,  # 'y'
    122: 20,  # 'z'
    123: CharacterCategory.SYMBOL,  # '{'
    124: CharacterCategory.SYMBOL,  # '|'
    125: CharacterCategory.SYMBOL,  # '}'
    126: CharacterCategory.SYMBOL,  # '~'
    127: CharacterCategory.CONTROL,  # '\x7f'
    128: 0,  # 'Ç'
    129: 24,  # 'ü'
    130: 0,  # 'é'
    131: 0,  # 'â'
    132: 27,  # 'ä'
    133: 0,  # 'à'
    134: 0,  # 'å'
    135: 0,  # 'ç'
    136: 0,  # 'ê'
    137: 0,  # 'ë'
    138: 0,  # 'è'
    139: 0,  # 'ï'
    140: 0,  # 'î'
    141: 0,  # 'ì'
    142: 59,  # 'Ä'
    143: 0,  # 'Å'
    144: 0,  # 'É'
    145: 0,  # 'æ'
    146: 0,  # 'Æ'
    147: 0,  # 'ô'
    148: 39,  # 'ö'
    149: 0,  # 'ò'
    150: 0,  # 'û'
    151: 0,  # 'ù'
    152: 0,  # 'ÿ'
    153: 58,  # 'Ö'
    154: 54,  # 'Ü'
    155: CharacterCategory.SYMBOL,  # '¢'
    156: CharacterCategory.SYMBOL,  # '£'
    157: CharacterCategory.SYMBOL,  # '¥'
    158: CharacterCategory.SYMBOL,  # '₧'
    159: 0,  # 'ƒ'
    160: 0,  # 'á'
    161: 0,  # 'í'
    162: 0,  # 'ó'
    163: 0,  # 'ú'
    164: 0,  # 'ñ'
    165: 0,  # 'Ñ'
    166: 0,  # 'ª'
    167: 0,  # 'º'
    168: CharacterCategory.SYMBOL,  # '¿'
    169: CharacterCategory.SYMBOL,  # '⌐'
    170: CharacterCategory.SYMBOL,  # '¬'
    171: CharacterCategory.DIGIT,  # '½'
    172: CharacterCategory.DIGIT,  # '¼'
    173: CharacterCategory.SYMBOL,  # '¡'
    174: CharacterCategory.SYMBOL,  # '«'
    175: CharacterCategory.SYMBOL,  # '»'
    176: CharacterCategory.SYMBOL,  # '░'
    177: CharacterCategory.SYMBOL,  # '▒'
    178: CharacterCategory.SYMBOL,  # '▓'
    179: CharacterCategory.SYMBOL,  # '│'
    180: CharacterCategory.SYMBOL,  # '┤'
    181: CharacterCategory.SYMBOL,  # '╡'
    182: CharacterCategory.SYMBOL,  # '╢'
    183: CharacterCategory.SYMBOL,  # '╖'
    184: CharacterCategory.SYMBOL,  # '╕'
    185: CharacterCategory.SYMBOL,  # '╣'
    186: CharacterCategory.SYMBOL,  # '║'
    187: CharacterCategory.SYMBOL,  # '╗'
    188: CharacterCategory.SYMBOL,  # '╝'
    189: CharacterCategory.SYMBOL,  # '╜'
    190: CharacterCategory.SYMBOL,  # '╛'
    191: CharacterCategory.SYMBOL,  # '┐'
    192: CharacterCategory.SYMBOL,  # '└'
    193: CharacterCategory.SYMBOL,  # '┴'
    194: CharacterCategory.SYMBOL,  # '┬'
    195: CharacterCategory.SYMBOL,  # '├'
    196: CharacterCategory.SYMBOL,  # '─'
    197: CharacterCategory.SYMBOL,  # '┼'
    198: CharacterCategory.SYMBOL,  # '╞'
    199: CharacterCategory.SYMBOL,  # '╟'
    200: CharacterCategory.SYMBOL,  # '╚'
    201: CharacterCategory.SYMBOL,  # '╔'
    202: CharacterCategory.SYMBOL,  # '╩'
    203: CharacterCategory.SYMBOL,  # '╦'
    204: CharacterCategory.SYMBOL,  # '╠'
    205: CharacterCategory.SYMBOL,  # '═'
    206: CharacterCategory.SYMBOL,  # '╬'
    207: CharacterCategory.SYMBOL,  # '╧'
    208: CharacterCategory.SYMBOL,  # '╨'
    209: CharacterCategory.SYMBOL,  # '╤'
    210: CharacterCategory.SYMBOL,  # '╥'
    211: CharacterCategory.SYMBOL,  # '╙'
    212: CharacterCategory.SYMBOL,  # '╘'
    213: CharacterCategory.SYMBOL,  # '╒'
    214: CharacterCategory.SYMBOL,  # '╓'
    215: CharacterCategory.SYMBOL,  # '╫'
    216: CharacterCategory.SYMBOL,  # '╪'
    217: CharacterCategory.SYMBOL,  # '┘'
    218: CharacterCategory.SYMBOL,  # '┌'
    219: CharacterCategory.SYMBOL,  # '█'
    220: CharacterCategory.SYMBOL,  # '▄'
    221: CharacterCategory.SYMBOL,  # '▌'
    222: CharacterCategory.SYMBOL,  # '▐'
    223: CharacterCategory.SYMBOL,  # '▀'
    224: 0,  # 'α'
    225: 48,  # 'ß'
    226: 0,  # 'Γ'
    227: 0,  # 'π'
    228: 0,  # 'Σ'
    229: 0,  # 'σ'
    230: 0,  # 'µ'
    231: 0,  # 'τ'
    232: 0,  # 'Φ'
    233: 0,  # 'Θ'
    234: 0,  # 'Ω'
    235: 0,  # 'δ'
    236: CharacterCategory.SYMBOL,  # '∞'
    237: 0,  # 'φ'
    238: 0,  # 'ε'
    239: CharacterCategory.SYMBOL,  # '∩'
    240: CharacterCategory.SYMBOL,  # '≡'
    241: CharacterCategory.SYMBOL,  # '±'
    242: CharacterCategory.SYMBOL,  # '≥'
    243: CharacterCategory.SYMBOL,  # '≤'
    244: CharacterCategory.SYMBOL,  # '⌠'
    245: CharacterCategory.SYMBOL,  # '⌡'
    246: CharacterCategory.SYMBOL,  # '÷'
    247: CharacterCategory.SYMBOL,  # '≈'
    248: CharacterCategory.SYMBOL,  # '°'
    249: CharacterCategory.SYMBOL,  # '∙'
    250: CharacterCategory.SYMBOL,  # '·'
    251: CharacterCategory.SYMBOL,  # '√'
    252: 0,  # 'ⁿ'
    253: CharacterCategory.DIGIT,  # '²'
    254: CharacterCategory.SYMBOL,  # '■'
    255: CharacterCategory.CONTROL,  # '\xa0'
}

CP437_GERMAN_MODEL = SingleByteCharSetModel(
    charset_name="CP437",
    language="German",
    char_to_order_map=CP437_GERMAN_CHAR_TO_ORDER,
    language_model=GERMAN_LANG_MODEL,
    typical_positive_ratio=0.9706854667596968,
    keep_ascii_letters=True,
    alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyzÄÖÜßäöüẞ",
)

CP500_GERMAN_CHAR_TO_ORDER = {
    0: CharacterCategory.CONTROL,  # '\x00'
    1: CharacterCategory.CONTROL,  # '\x01'
    2: CharacterCategory.CONTROL,  # '\x02'
    3: CharacterCategory.CONTROL,  # '\x03'
    4: CharacterCategory.CONTROL,  # '\x9c'
    5: CharacterCategory.CONTROL,  # '\t'
    6: CharacterCategory.CONTROL,  # '\x86'
    7: CharacterCategory.CONTROL,  # '\x7f'
    8: CharacterCategory.CONTROL,  # '\x97'
    9: CharacterCategory.CONTROL,  # '\x8d'
    10: CharacterCategory.CONTROL,  # '\x8e'
    11: CharacterCategory.CONTROL,  # '\x0b'
    12: CharacterCategory.CONTROL,  # '\x0c'
    13: CharacterCategory.LINE_BREAK,  # '\r'
    14: CharacterCategory.CONTROL,  # '\x0e'
    15: CharacterCategory.CONTROL,  # '\x0f'
    16: CharacterCategory.CONTROL,  # '\x10'
    17: CharacterCategory.CONTROL,  # '\x11'
    18: CharacterCategory.CONTROL,  # '\x12'
    19: CharacterCategory.CONTROL,  # '\x13'
    20: CharacterCategory.CONTROL,  # '\x9d'
    21: CharacterCategory.CONTROL,  # '\x85'
    22: CharacterCategory.CONTROL,  # '\x08'
    23: CharacterCategory.CONTROL,  # '\x87'
    24: CharacterCategory.CONTROL,  # '\x18'
    25: CharacterCategory.CONTROL,  # '\x19'
    26: CharacterCategory.CONTROL,  # '\x92'
    27: CharacterCategory.CONTROL,  # '\x8f'
    28: CharacterCategory.CONTROL,  # '\x1c'
    29: CharacterCategory.CONTROL,  # '\x1d'
    30: CharacterCategory.CONTROL,  # '\x1e'
    31: CharacterCategory.CONTROL,  # '\x1f'
    32: CharacterCategory.CONTROL,  # '\x80'
    33: CharacterCategory.CONTROL,  # '\x81'
    34: CharacterCategory.CONTROL,  # '\x82'
    35: CharacterCategory.CONTROL,  # '\x83'
    36: CharacterCategory.CONTROL,  # '\x84'
    37: CharacterCategory.LINE_BREAK,  # '\n'
    38: CharacterCategory.CONTROL,  # '\x17'
    39: CharacterCategory.CONTROL,  # '\x1b'
    40: CharacterCategory.CONTROL,  # '\x88'
    41: CharacterCategory.CONTROL,  # '\x89'
    42: CharacterCategory.CONTROL,  # '\x8a'
    43: CharacterCategory.CONTROL,  # '\x8b'
    44: CharacterCategory.CONTROL,  # '\x8c'
    45: CharacterCategory.CONTROL,  # '\x05'
    46: CharacterCategory.CONTROL,  # '\x06'
    47: CharacterCategory.CONTROL,  # '\x07'
    48: CharacterCategory.CONTROL,  # '\x90'
    49: CharacterCategory.CONTROL,  # '\x91'
    50: CharacterCategory.CONTROL,  # '\x16'
    51: CharacterCategory.CONTROL,  # '\x93'
    52: CharacterCategory.CONTROL,  # '\x94'
    53: CharacterCategory.CONTROL,  # '\x95'
    54: CharacterCategory.CONTROL,  # '\x96'
    55: CharacterCategory.CONTROL,  # '\x04'
    56: CharacterCategory.CONTROL,  # '\x98'
    57: CharacterCategory.CONTROL,  # '\x99'
    58: CharacterCategory.CONTROL,  # '\x9a'
    59: CharacterCategory.CONTROL,  # '\x9b'
    60: CharacterCategory.CONTROL,  # '\x14'
    61: CharacterCategory.CONTROL,  # '\x15'
    62: CharacterCategory.CONTROL,  # '\x9e'
    63: CharacterCategory.CONTROL,  # '\x1a'
    64: CharacterCategory.CONTROL,  # ' '
    65: CharacterCategory.CONTROL,  # '\xa0'
    66: 0,  # 'â'
    67: 27,  # 'ä'
    68: 0,  # 'à'
    69: 0,  # 'á'
    70: 0,  # 'ã'
    71: 0,  # 'å'
    72: 0,  # 'ç'
    73: 0,  # 'ñ'
    74: CharacterCategory.SYMBOL,  # '['
    75: CharacterCategory.SYMBOL,  # '.'
    76: CharacterCategory.SYMBOL,  # '<'
    77: CharacterCategory.SYMBOL,  # '('
    78: CharacterCategory.SYMBOL,  # '+'
    79: CharacterCategory.SYMBOL,  # '!'
    80: CharacterCategory.SYMBOL,  # '&'
    81: 0,  # 'é'
    82: 0,  # 'ê'
    83: 0,  # 'ë'
    84: 0,  # 'è'
    85: 0,  # 'í'
    86: 0,  # 'î'
    87: 0,  # 'ï'
    88: 0,  # 'ì'
    89: 48,  # 'ß'
    90: CharacterCategory.SYMBOL,  # ']'
    91: CharacterCategory.SYMBOL,  # '$'
    92: CharacterCategory.SYMBOL,  # '*'
    93: CharacterCategory.SYMBOL,  # ')'
    94: CharacterCategory.SYMBOL,  # ';'
    95: CharacterCategory.SYMBOL,  # '^'
    96: CharacterCategory.SYMBOL,  # '-'
    97: CharacterCategory.SYMBOL,  # '/'
    98: 0,  # 'Â'
    99: 59,  # 'Ä'
    100: 0,  # 'À'
    101: 0,  # 'Á'
    102: 0,  # 'Ã'
    103: 0,  # 'Å'
    104: 0,  # 'Ç'
    105: 0,  # 'Ñ'
    106: CharacterCategory.SYMBOL,  # '¦'
    107: CharacterCategory.SYMBOL,  # ','
    108: CharacterCategory.SYMBOL,  # '%'
    109: CharacterCategory.SYMBOL,  # '_'
    110: CharacterCategory.SYMBOL,  # '>'
    111: CharacterCategory.SYMBOL,  # '?'
    112: 0,  # 'ø'
    113: 0,  # 'É'
    114: 0,  # 'Ê'
    115: 0,  # 'Ë'
    116: 0,  # 'È'
    117: 0,  # 'Í'
    118: 0,  # 'Î'
    119: 0,  # 'Ï'
    120: 0,  # 'Ì'
    121: CharacterCategory.SYMBOL,  # '`'
    122: CharacterCategory.SYMBOL,  # ':'
    123: CharacterCategory.SYMBOL,  # '#'
    124: CharacterCategory.SYMBOL,  # '@'
    125: CharacterCategory.SYMBOL,  # "'"
    126: CharacterCategory.SYMBOL,  # '='
    127: CharacterCategory.SYMBOL,  # '"'
    128: 0,  # 'Ø'
    129: 7,  # 'a'
    130: 16,  # 'b'
    131: 13,  # 'c'
    132: 8,  # 'd'
    133: 1,  # 'e'
    134: 17,  # 'f'
    135: 14,  # 'g'
    136: 9,  # 'h'
    137: 3,  # 'i'
    138: CharacterCategory.SYMBOL,  # '«'
    139: CharacterCategory.SYMBOL,  # '»'
    140: 0,  # 'ð'
    141: 0,  # 'ý'
    142: 0,  # 'þ'
    143: CharacterCategory.SYMBOL,  # '±'
    144: CharacterCategory.SYMBOL,  # '°'
    145: 50,  # 'j'
    146: 18,  # 'k'
    147: 11,  # 'l'
    148: 15,  # 'm'
    149: 2,  # 'n'
    150: 12,  # 'o'
    151: 21,  # 'p'
    152: 53,  # 'q'
    153: 4,  # 'r'
    154: 0,  # 'ª'
    155: 0,  # 'º'
    156: 0,  # 'æ'
    157: CharacterCategory.SYMBOL,  # '¸'
    158: 0,  # 'Æ'
    159: CharacterCategory.SYMBOL,  # '¤'
    160: 0,  # 'µ'
    161: CharacterCategory.SYMBOL,  # '~'
    162: 6,  # 's'
    163: 5,  # 't'
    164: 10,  # 'u'
    165: 23,  # 'v'
    166: 19,  # 'w'
    167: 52,  # 'x'
    168: 44,  # 'y'
    169: 20,  # 'z'
    170: CharacterCategory.SYMBOL,  # '¡'
    171: CharacterCategory.SYMBOL,  # '¿'
    172: 0,  # 'Ð'
    173: 0,  # 'Ý'
    174: 0,  # 'Þ'
    175: CharacterCategory.SYMBOL,  # '®'
    176: CharacterCategory.SYMBOL,  # '¢'
    177: CharacterCategory.SYMBOL,  # '£'
    178: CharacterCategory.SYMBOL,  # '¥'
    179: CharacterCategory.SYMBOL,  # '·'
    180: CharacterCategory.SYMBOL,  # '©'
    181: CharacterCategory.SYMBOL,  # '§'
    182: CharacterCategory.SYMBOL,  # '¶'
    183: CharacterCategory.DIGIT,  # '¼'
    184: CharacterCategory.DIGIT,  # '½'
    185: CharacterCategory.DIGIT,  # '¾'
    186: CharacterCategory.SYMBOL,  # '¬'
    187: CharacterCategory.SYMBOL,  # '|'
    188: CharacterCategory.SYMBOL,  # '¯'
    189: CharacterCategory.SYMBOL,  # '¨'
    190: CharacterCategory.SYMBOL,  # '´'
    191: CharacterCategory.SYMBOL,  # '×'
    192: CharacterCategory.SYMBOL,  # '{'
    193: 25,  # 'A'
    194: 28,  # 'B'
    195: 45,  # 'C'
    196: 26,  # 'D'
    197: 30,  # 'E'
    198: 33,  # 'F'
    199: 35,  # 'G'
    200: 38,  # 'H'
    201: 37,  # 'I'
    202: CharacterCategory.CONTROL,  # '\xad'
    203: 0,  # 'ô'
    204: 39,  # 'ö'
    205: 0,  # 'ò'
    206: 0,  # 'ó'
    207: 0,  # 'õ'
    208: CharacterCategory.SYMBOL,  # '}'
    209: 51,  # 'J'
    210: 32,  # 'K'
    211: 41,  # 'L'
    212: 29,  # 'M'
    213: 43,  # 'N'
    214: 49,  # 'O'
    215: 34,  # 'P'
    216: 55,  # 'Q'
    217: 40,  # 'R'
    218: CharacterCategory.DIGIT,  # '¹'
    219: 0,  # 'û'
    220: 24,  # 'ü'
    221: 0,  # 'ù'
    222: 0,  # 'ú'
    223: 0,  # 'ÿ'
    224: CharacterCategory.SYMBOL,  # '\\'
    225: CharacterCategory.SYMBOL,  # '÷'
    226: 22,  # 'S'
    227: 36,  # 'T'
    228: 46,  # 'U'
    229: 42,  # 'V'
    230: 31,  # 'W'
    231: 57,  # 'X'
    232: 56,  # 'Y'
    233: 47,  # 'Z'
    234: CharacterCategory.DIGIT,  # '²'
    235: 0,  # 'Ô'
    236: 58,  # 'Ö'
    237: 0,  # 'Ò'
    238: 0,  # 'Ó'
    239: 0,  # 'Õ'
    240: CharacterCategory.DIGIT,  # '0'
    241: CharacterCategory.DIGIT,  # '1'
    242: CharacterCategory.DIGIT,  # '2'
    243: CharacterCategory.DIGIT,  # '3'
    244: CharacterCategory.DIGIT,  # '4'
    245: CharacterCategory.DIGIT,  # '5'
    246: CharacterCategory.DIGIT,  # '6'
    247: CharacterCategory.DIGIT,  # '7'
    248: CharacterCategory.DIGIT,  # '8'
    249: CharacterCategory.DIGIT,  # '9'
    250: CharacterCategory.DIGIT,  # '³'
    251: 0,  # 'Û'
    252: 54,  # 'Ü'
    253: 0,  # 'Ù'
    254: 0,  # 'Ú'
    255: CharacterCategory.CONTROL,  # '\x9f'
}

CP500_GERMAN_MODEL = SingleByteCharSetModel(
    charset_name="CP500",
    language="German",
    char_to_order_map=CP500_GERMAN_CHAR_TO_ORDER,
    language_model=GERMAN_LANG_MODEL,
    typical_positive_ratio=0.9706854667596968,
    keep_ascii_letters=True,
    alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyzÄÖÜßäöüẞ",
)

CP850_GERMAN_CHAR_TO_ORDER = {
    0: CharacterCategory.CONTROL,  # '\x00'
    1: CharacterCategory.CONTROL,  # '\x01'
    2: CharacterCategory.CONTROL,  # '\x02'
    3: CharacterCategory.CONTROL,  # '\x03'
    4: CharacterCategory.CONTROL,  # '\x04'
    5: CharacterCategory.CONTROL,  # '\x05'
    6: CharacterCategory.CONTROL,  # '\x06'
    7: CharacterCategory.CONTROL,  # '\x07'
    8: CharacterCategory.CONTROL,  # '\x08'
    9: CharacterCategory.CONTROL,  # '\t'
    10: CharacterCategory.LINE_BREAK,  # '\n'
    11: CharacterCategory.CONTROL,  # '\x0b'
    12: CharacterCategory.CONTROL,  # '\x0c'
    13: CharacterCategory.LINE_BREAK,  # '\r'
    14: CharacterCategory.CONTROL,  # '\x0e'
    15: CharacterCategory.CONTROL,  # '\x0f'
    16: CharacterCategory.CONTROL,  # '\x10'
    17: CharacterCategory.CONTROL,  # '\x11'
    18: CharacterCategory.CONTROL,  # '\x12'
    19: CharacterCategory.CONTROL,  # '\x13'
    20: CharacterCategory.CONTROL,  # '\x14'
    21: CharacterCategory.CONTROL,  # '\x15'
    22: CharacterCategory.CONTROL,  # '\x16'
    23: CharacterCategory.CONTROL,  # '\x17'
    24: CharacterCategory.CONTROL,  # '\x18'
    25: CharacterCategory.CONTROL,  # '\x19'
    26: CharacterCategory.CONTROL,  # '\x1a'
    27: CharacterCategory.CONTROL,  # '\x1b'
    28: CharacterCategory.CONTROL,  # '\x1c'
    29: CharacterCategory.CONTROL,  # '\x1d'
    30: CharacterCategory.CONTROL,  # '\x1e'
    31: CharacterCategory.CONTROL,  # '\x1f'
    32: CharacterCategory.CONTROL,  # ' '
    33: CharacterCategory.SYMBOL,  # '!'
    34: CharacterCategory.SYMBOL,  # '"'
    35: CharacterCategory.SYMBOL,  # '#'
    36: CharacterCategory.SYMBOL,  # '$'
    37: CharacterCategory.SYMBOL,  # '%'
    38: CharacterCategory.SYMBOL,  # '&'
    39: CharacterCategory.SYMBOL,  # "'"
    40: CharacterCategory.SYMBOL,  # '('
    41: CharacterCategory.SYMBOL,  # ')'
    42: CharacterCategory.SYMBOL,  # '*'
    43: CharacterCategory.SYMBOL,  # '+'
    44: CharacterCategory.SYMBOL,  # ','
    45: CharacterCategory.SYMBOL,  # '-'
    46: CharacterCategory.SYMBOL,  # '.'
    47: CharacterCategory.SYMBOL,  # '/'
    48: CharacterCategory.DIGIT,  # '0'
    49: CharacterCategory.DIGIT,  # '1'
    50: CharacterCategory.DIGIT,  # '2'
    51: CharacterCategory.DIGIT,  # '3'
    52: CharacterCategory.DIGIT,  # '4'
    53: CharacterCategory.DIGIT,  # '5'
    54: CharacterCategory.DIGIT,  # '6'
    55: CharacterCategory.DIGIT,  # '7'
    56: CharacterCategory.DIGIT,  # '8'
    57: CharacterCategory.DIGIT,  # '9'
    58: CharacterCategory.SYMBOL,  # ':'
    59: CharacterCategory.SYMBOL,  # ';'
    60: CharacterCategory.SYMBOL,  # '<'
    61: CharacterCategory.SYMBOL,  # '='
    62: CharacterCategory.SYMBOL,  # '>'
    63: CharacterCategory.SYMBOL,  # '?'
    64: CharacterCategory.SYMBOL,  # '@'
    65: 25,  # 'A'
    66: 28,  # 'B'
    67: 45,  # 'C'
    68: 26,  # 'D'
    69: 30,  # 'E'
    70: 33,  # 'F'
    71: 35,  # 'G'
    72: 38,  # 'H'
    73: 37,  # 'I'
    74: 51,  # 'J'
    75: 32,  # 'K'
    76: 41,  # 'L'
    77: 29,  # 'M'
    78: 43,  # 'N'
    79: 49,  # 'O'
    80: 34,  # 'P'
    81: 55,  # 'Q'
    82: 40,  # 'R'
    83: 22,  # 'S'
    84: 36,  # 'T'
    85: 46,  # 'U'
    86: 42,  # 'V'
    87: 31,  # 'W'
    88: 57,  # 'X'
    89: 56,  # 'Y'
    90: 47,  # 'Z'
    91: CharacterCategory.SYMBOL,  # '['
    92: CharacterCategory.SYMBOL,  # '\\'
    93: CharacterCategory.SYMBOL,  # ']'
    94: CharacterCategory.SYMBOL,  # '^'
    95: CharacterCategory.SYMBOL,  # '_'
    96: CharacterCategory.SYMBOL,  # '`'
    97: 7,  # 'a'
    98: 16,  # 'b'
    99: 13,  # 'c'
    100: 8,  # 'd'
    101: 1,  # 'e'
    102: 17,  # 'f'
    103: 14,  # 'g'
    104: 9,  # 'h'
    105: 3,  # 'i'
    106: 50,  # 'j'
    107: 18,  # 'k'
    108: 11,  # 'l'
    109: 15,  # 'm'
    110: 2,  # 'n'
    111: 12,  # 'o'
    112: 21,  # 'p'
    113: 53,  # 'q'
    114: 4,  # 'r'
    115: 6,  # 's'
    116: 5,  # 't'
    117: 10,  # 'u'
    118: 23,  # 'v'
    119: 19,  # 'w'
    120: 52,  # 'x'
    121: 44,  # 'y'
    122: 20,  # 'z'
    123: CharacterCategory.SYMBOL,  # '{'
    124: CharacterCategory.SYMBOL,  # '|'
    125: CharacterCategory.SYMBOL,  # '}'
    126: CharacterCategory.SYMBOL,  # '~'
    127: CharacterCategory.CONTROL,  # '\x7f'
    128: 0,  # 'Ç'
    129: 24,  # 'ü'
    130: 0,  # 'é'
    131: 0,  # 'â'
    132: 27,  # 'ä'
    133: 0,  # 'à'
    134: 0,  # 'å'
    135: 0,  # 'ç'
    136: 0,  # 'ê'
    137: 0,  # 'ë'
    138: 0,  # 'è'
    139: 0,  # 'ï'
    140: 0,  # 'î'
    141: 0,  # 'ì'
    142: 59,  # 'Ä'
    143: 0,  # 'Å'
    144: 0,  # 'É'
    145: 0,  # 'æ'
    146: 0,  # 'Æ'
    147: 0,  # 'ô'
    148: 39,  # 'ö'
    149: 0,  # 'ò'
    150: 0,  # 'û'
    151: 0,  # 'ù'
    152: 0,  # 'ÿ'
    153: 58,  # 'Ö'
    154: 54,  # 'Ü'
    155: 0,  # 'ø'
    156: CharacterCategory.SYMBOL,  # '£'
    157: 0,  # 'Ø'
    158: CharacterCategory.SYMBOL,  # '×'
    159: 0,  # 'ƒ'
    160: 0,  # 'á'
    161: 0,  # 'í'
    162: 0,  # 'ó'
    163: 0,  # 'ú'
    164: 0,  # 'ñ'
    165: 0,  # 'Ñ'
    166: 0,  # 'ª'
    167: 0,  # 'º'
    168: CharacterCategory.SYMBOL,  # '¿'
    169: CharacterCategory.SYMBOL,  # '®'
    170: CharacterCategory.SYMBOL,  # '¬'
    171: CharacterCategory.DIGIT,  # '½'
    172: CharacterCategory.DIGIT,  # '¼'
    173: CharacterCategory.SYMBOL,  # '¡'
    174: CharacterCategory.SYMBOL,  # '«'
    175: CharacterCategory.SYMBOL,  # '»'
    176: CharacterCategory.SYMBOL,  # '░'
    177: CharacterCategory.SYMBOL,  # '▒'
    178: CharacterCategory.SYMBOL,  # '▓'
    179: CharacterCategory.SYMBOL,  # '│'
    180: CharacterCategory.SYMBOL,  # '┤'
    181: 0,  # 'Á'
    182: 0,  # 'Â'
    183: 0,  # 'À'
    184: CharacterCategory.SYMBOL,  # '©'
    185: CharacterCategory.SYMBOL,  # '╣'
    186: CharacterCategory.SYMBOL,  # '║'
    187: CharacterCategory.SYMBOL,  # '╗'
    188: CharacterCategory.SYMBOL,  # '╝'
    189: CharacterCategory.SYMBOL,  # '¢'
    190: CharacterCategory.SYMBOL,  # '¥'
    191: CharacterCategory.SYMBOL,  # '┐'
    192: CharacterCategory.SYMBOL,  # '└'
    193: CharacterCategory.SYMBOL,  # '┴'
    194: CharacterCategory.SYMBOL,  # '┬'
    195: CharacterCategory.SYMBOL,  # '├'
    196: CharacterCategory.SYMBOL,  # '─'
    197: CharacterCategory.SYMBOL,  # '┼'
    198: 0,  # 'ã'
    199: 0,  # 'Ã'
    200: CharacterCategory.SYMBOL,  # '╚'
    201: CharacterCategory.SYMBOL,  # '╔'
    202: CharacterCategory.SYMBOL,  # '╩'
    203: CharacterCategory.SYMBOL,  # '╦'
    204: CharacterCategory.SYMBOL,  # '╠'
    205: CharacterCategory.SYMBOL,  # '═'
    206: CharacterCategory.SYMBOL,  # '╬'
    207: CharacterCategory.SYMBOL,  # '¤'
    208: 0,  # 'ð'
    209: 0,  # 'Ð'
    210: 0,  # 'Ê'
    211: 0,  # 'Ë'
    212: 0,  # 'È'
    213: 0,  # 'ı'
    214: 0,  # 'Í'
    215: 0,  # 'Î'
    216: 0,  # 'Ï'
    217: CharacterCategory.SYMBOL,  # '┘'
    218: CharacterCategory.SYMBOL,  # '┌'
    219: CharacterCategory.SYMBOL,  # '█'
    220: CharacterCategory.SYMBOL,  # '▄'
    221: CharacterCategory.SYMBOL,  # '¦'
    222: 0,  # 'Ì'
    223: CharacterCategory.SYMBOL,  # '▀'
    224: 0,  # 'Ó'
    225: 48,  # 'ß'
    226: 0,  # 'Ô'
    227: 0,  # 'Ò'
    228: 0,  # 'õ'
    229: 0,  # 'Õ'
    230: 0,  # 'µ'
    231: 0,  # 'þ'
    232: 0,  # 'Þ'
    233: 0,  # 'Ú'
    234: 0,  # 'Û'
    235: 0,  # 'Ù'
    236: 0,  # 'ý'
    237: 0,  # 'Ý'
    238: CharacterCategory.SYMBOL,  # '¯'
    239: CharacterCategory.SYMBOL,  # '´'
    240: CharacterCategory.CONTROL,  # '\xad'
    241: CharacterCategory.SYMBOL,  # '±'
    242: CharacterCategory.SYMBOL,  # '‗'
    243: CharacterCategory.DIGIT,  # '¾'
    244: CharacterCategory.SYMBOL,  # '¶'
    245: CharacterCategory.SYMBOL,  # '§'
    246: CharacterCategory.SYMBOL,  # '÷'
    247: CharacterCategory.SYMBOL,  # '¸'
    248: CharacterCategory.SYMBOL,  # '°'
    249: CharacterCategory.SYMBOL,  # '¨'
    250: CharacterCategory.SYMBOL,  # '·'
    251: CharacterCategory.DIGIT,  # '¹'
    252: CharacterCategory.DIGIT,  # '³'
    253: CharacterCategory.DIGIT,  # '²'
    254: CharacterCategory.SYMBOL,  # '■'
    255: CharacterCategory.CONTROL,  # '\xa0'
}

CP850_GERMAN_MODEL = SingleByteCharSetModel(
    charset_name="CP850",
    language="German",
    char_to_order_map=CP850_GERMAN_CHAR_TO_ORDER,
    language_model=GERMAN_LANG_MODEL,
    typical_positive_ratio=0.9706854667596968,
    keep_ascii_letters=True,
    alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyzÄÖÜßäöüẞ",
)

CP858_GERMAN_CHAR_TO_ORDER = {
    0: CharacterCategory.CONTROL,  # '\x00'
    1: CharacterCategory.CONTROL,  # '\x01'
    2: CharacterCategory.CONTROL,  # '\x02'
    3: CharacterCategory.CONTROL,  # '\x03'
    4: CharacterCategory.CONTROL,  # '\x04'
    5: CharacterCategory.CONTROL,  # '\x05'
    6: CharacterCategory.CONTROL,  # '\x06'
    7: CharacterCategory.CONTROL,  # '\x07'
    8: CharacterCategory.CONTROL,  # '\x08'
    9: CharacterCategory.CONTROL,  # '\t'
    10: CharacterCategory.LINE_BREAK,  # '\n'
    11: CharacterCategory.CONTROL,  # '\x0b'
    12: CharacterCategory.CONTROL,  # '\x0c'
    13: CharacterCategory.LINE_BREAK,  # '\r'
    14: CharacterCategory.CONTROL,  # '\x0e'
    15: CharacterCategory.CONTROL,  # '\x0f'
    16: CharacterCategory.CONTROL,  # '\x10'
    17: CharacterCategory.CONTROL,  # '\x11'
    18: CharacterCategory.CONTROL,  # '\x12'
    19: CharacterCategory.CONTROL,  # '\x13'
    20: CharacterCategory.CONTROL,  # '\x14'
    21: CharacterCategory.CONTROL,  # '\x15'
    22: CharacterCategory.CONTROL,  # '\x16'
    23: CharacterCategory.CONTROL,  # '\x17'
    24: CharacterCategory.CONTROL,  # '\x18'
    25: CharacterCategory.CONTROL,  # '\x19'
    26: CharacterCategory.CONTROL,  # '\x1a'
    27: CharacterCategory.CONTROL,  # '\x1b'
    28: CharacterCategory.CONTROL,  # '\x1c'
    29: CharacterCategory.CONTROL,  # '\x1d'
    30: CharacterCategory.CONTROL,  # '\x1e'
    31: CharacterCategory.CONTROL,  # '\x1f'
    32: CharacterCategory.CONTROL,  # ' '
    33: CharacterCategory.SYMBOL,  # '!'
    34: CharacterCategory.SYMBOL,  # '"'
    35: CharacterCategory.SYMBOL,  # '#'
    36: CharacterCategory.SYMBOL,  # '$'
    37: CharacterCategory.SYMBOL,  # '%'
    38: CharacterCategory.SYMBOL,  # '&'
    39: CharacterCategory.SYMBOL,  # "'"
    40: CharacterCategory.SYMBOL,  # '('
    41: CharacterCategory.SYMBOL,  # ')'
    42: CharacterCategory.SYMBOL,  # '*'
    43: CharacterCategory.SYMBOL,  # '+'
    44: CharacterCategory.SYMBOL,  # ','
    45: CharacterCategory.SYMBOL,  # '-'
    46: CharacterCategory.SYMBOL,  # '.'
    47: CharacterCategory.SYMBOL,  # '/'
    48: CharacterCategory.DIGIT,  # '0'
    49: CharacterCategory.DIGIT,  # '1'
    50: CharacterCategory.DIGIT,  # '2'
    51: CharacterCategory.DIGIT,  # '3'
    52: CharacterCategory.DIGIT,  # '4'
    53: CharacterCategory.DIGIT,  # '5'
    54: CharacterCategory.DIGIT,  # '6'
    55: CharacterCategory.DIGIT,  # '7'
    56: CharacterCategory.DIGIT,  # '8'
    57: CharacterCategory.DIGIT,  # '9'
    58: CharacterCategory.SYMBOL,  # ':'
    59: CharacterCategory.SYMBOL,  # ';'
    60: CharacterCategory.SYMBOL,  # '<'
    61: CharacterCategory.SYMBOL,  # '='
    62: CharacterCategory.SYMBOL,  # '>'
    63: CharacterCategory.SYMBOL,  # '?'
    64: CharacterCategory.SYMBOL,  # '@'
    65: 25,  # 'A'
    66: 28,  # 'B'
    67: 45,  # 'C'
    68: 26,  # 'D'
    69: 30,  # 'E'
    70: 33,  # 'F'
    71: 35,  # 'G'
    72: 38,  # 'H'
    73: 37,  # 'I'
    74: 51,  # 'J'
    75: 32,  # 'K'
    76: 41,  # 'L'
    77: 29,  # 'M'
    78: 43,  # 'N'
    79: 49,  # 'O'
    80: 34,  # 'P'
    81: 55,  # 'Q'
    82: 40,  # 'R'
    83: 22,  # 'S'
    84: 36,  # 'T'
    85: 46,  # 'U'
    86: 42,  # 'V'
    87: 31,  # 'W'
    88: 57,  # 'X'
    89: 56,  # 'Y'
    90: 47,  # 'Z'
    91: CharacterCategory.SYMBOL,  # '['
    92: CharacterCategory.SYMBOL,  # '\\'
    93: CharacterCategory.SYMBOL,  # ']'
    94: CharacterCategory.SYMBOL,  # '^'
    95: CharacterCategory.SYMBOL,  # '_'
    96: CharacterCategory.SYMBOL,  # '`'
    97: 7,  # 'a'
    98: 16,  # 'b'
    99: 13,  # 'c'
    100: 8,  # 'd'
    101: 1,  # 'e'
    102: 17,  # 'f'
    103: 14,  # 'g'
    104: 9,  # 'h'
    105: 3,  # 'i'
    106: 50,  # 'j'
    107: 18,  # 'k'
    108: 11,  # 'l'
    109: 15,  # 'm'
    110: 2,  # 'n'
    111: 12,  # 'o'
    112: 21,  # 'p'
    113: 53,  # 'q'
    114: 4,  # 'r'
    115: 6,  # 's'
    116: 5,  # 't'
    117: 10,  # 'u'
    118: 23,  # 'v'
    119: 19,  # 'w'
    120: 52,  # 'x'
    121: 44,  # 'y'
    122: 20,  # 'z'
    123: CharacterCategory.SYMBOL,  # '{'
    124: CharacterCategory.SYMBOL,  # '|'
    125: CharacterCategory.SYMBOL,  # '}'
    126: CharacterCategory.SYMBOL,  # '~'
    127: CharacterCategory.CONTROL,  # '\x7f'
    128: 0,  # 'Ç'
    129: 24,  # 'ü'
    130: 0,  # 'é'
    131: 0,  # 'â'
    132: 27,  # 'ä'
    133: 0,  # 'à'
    134: 0,  # 'å'
    135: 0,  # 'ç'
    136: 0,  # 'ê'
    137: 0,  # 'ë'
    138: 0,  # 'è'
    139: 0,  # 'ï'
    140: 0,  # 'î'
    141: 0,  # 'ì'
    142: 59,  # 'Ä'
    143: 0,  # 'Å'
    144: 0,  # 'É'
    145: 0,  # 'æ'
    146: 0,  # 'Æ'
    147: 0,  # 'ô'
    148: 39,  # 'ö'
    149: 0,  # 'ò'
    150: 0,  # 'û'
    151: 0,  # 'ù'
    152: 0,  # 'ÿ'
    153: 58,  # 'Ö'
    154: 54,  # 'Ü'
    155: 0,  # 'ø'
    156: CharacterCategory.SYMBOL,  # '£'
    157: 0,  # 'Ø'
    158: CharacterCategory.SYMBOL,  # '×'
    159: 0,  # 'ƒ'
    160: 0,  # 'á'
    161: 0,  # 'í'
    162: 0,  # 'ó'
    163: 0,  # 'ú'
    164: 0,  # 'ñ'
    165: 0,  # 'Ñ'
    166: 0,  # 'ª'
    167: 0,  # 'º'
    168: CharacterCategory.SYMBOL,  # '¿'
    169: CharacterCategory.SYMBOL,  # '®'
    170: CharacterCategory.SYMBOL,  # '¬'
    171: CharacterCategory.DIGIT,  # '½'
    172: CharacterCategory.DIGIT,  # '¼'
    173: CharacterCategory.SYMBOL,  # '¡'
    174: CharacterCategory.SYMBOL,  # '«'
    175: CharacterCategory.SYMBOL,  # '»'
    176: CharacterCategory.SYMBOL,  # '░'
    177: CharacterCategory.SYMBOL,  # '▒'
    178: CharacterCategory.SYMBOL,  # '▓'
    179: CharacterCategory.SYMBOL,  # '│'
    180: CharacterCategory.SYMBOL,  # '┤'
    181: 0,  # 'Á'
    182: 0,  # 'Â'
    183: 0,  # 'À'
    184: CharacterCategory.SYMBOL,  # '©'
    185: CharacterCategory.SYMBOL,  # '╣'
    186: CharacterCategory.SYMBOL,  # '║'
    187: CharacterCategory.SYMBOL,  # '╗'
    188: CharacterCategory.SYMBOL,  # '╝'
    189: CharacterCategory.SYMBOL,  # '¢'
    190: CharacterCategory.SYMBOL,  # '¥'
    191: CharacterCategory.SYMBOL,  # '┐'
    192: CharacterCategory.SYMBOL,  # '└'
    193: CharacterCategory.SYMBOL,  # '┴'
    194: CharacterCategory.SYMBOL,  # '┬'
    195: CharacterCategory.SYMBOL,  # '├'
    196: CharacterCategory.SYMBOL,  # '─'
    197: CharacterCategory.SYMBOL,  # '┼'
    198: 0,  # 'ã'
    199: 0,  # 'Ã'
    200: CharacterCategory.SYMBOL,  # '╚'
    201: CharacterCategory.SYMBOL,  # '╔'
    202: CharacterCategory.SYMBOL,  # '╩'
    203: CharacterCategory.SYMBOL,  # '╦'
    204: CharacterCategory.SYMBOL,  # '╠'
    205: CharacterCategory.SYMBOL,  # '═'
    206: CharacterCategory.SYMBOL,  # '╬'
    207: CharacterCategory.SYMBOL,  # '¤'
    208: 0,  # 'ð'
    209: 0,  # 'Ð'
    210: 0,  # 'Ê'
    211: 0,  # 'Ë'
    212: 0,  # 'È'
    213: CharacterCategory.SYMBOL,  # '€'
    214: 0,  # 'Í'
    215: 0,  # 'Î'
    216: 0,  # 'Ï'
    217: CharacterCategory.SYMBOL,  # '┘'
    218: CharacterCategory.SYMBOL,  # '┌'
    219: CharacterCategory.SYMBOL,  # '█'
    220: CharacterCategory.SYMBOL,  # '▄'
    221: CharacterCategory.SYMBOL,  # '¦'
    222: 0,  # 'Ì'
    223: CharacterCategory.SYMBOL,  # '▀'
    224: 0,  # 'Ó'
    225: 48,  # 'ß'
    226: 0,  # 'Ô'
    227: 0,  # 'Ò'
    228: 0,  # 'õ'
    229: 0,  # 'Õ'
    230: 0,  # 'µ'
    231: 0,  # 'þ'
    232: 0,  # 'Þ'
    233: 0,  # 'Ú'
    234: 0,  # 'Û'
    235: 0,  # 'Ù'
    236: 0,  # 'ý'
    237: 0,  # 'Ý'
    238: CharacterCategory.SYMBOL,  # '¯'
    239: CharacterCategory.SYMBOL,  # '´'
    240: CharacterCategory.CONTROL,  # '\xad'
    241: CharacterCategory.SYMBOL,  # '±'
    242: CharacterCategory.SYMBOL,  # '‗'
    243: CharacterCategory.DIGIT,  # '¾'
    244: CharacterCategory.SYMBOL,  # '¶'
    245: CharacterCategory.SYMBOL,  # '§'
    246: CharacterCategory.SYMBOL,  # '÷'
    247: CharacterCategory.SYMBOL,  # '¸'
    248: CharacterCategory.SYMBOL,  # '°'
    249: CharacterCategory.SYMBOL,  # '¨'
    250: CharacterCategory.SYMBOL,  # '·'
    251: CharacterCategory.DIGIT,  # '¹'
    252: CharacterCategory.DIGIT,  # '³'
    253: CharacterCategory.DIGIT,  # '²'
    254: CharacterCategory.SYMBOL,  # '■'
    255: CharacterCategory.CONTROL,  # '\xa0'
}

CP858_GERMAN_MODEL = SingleByteCharSetModel(
    charset_name="CP858",
    language="German",
    char_to_order_map=CP858_GERMAN_CHAR_TO_ORDER,
    language_model=GERMAN_LANG_MODEL,
    typical_positive_ratio=0.9706854667596968,
    keep_ascii_letters=True,
    alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyzÄÖÜßäöüẞ",
)

ISO_8859_1_GERMAN_CHAR_TO_ORDER = {
    0: CharacterCategory.CONTROL,  # '\x00'
    1: CharacterCategory.CONTROL,  # '\x01'
    2: CharacterCategory.CONTROL,  # '\x02'
    3: CharacterCategory.CONTROL,  # '\x03'
    4: CharacterCategory.CONTROL,  # '\x04'
    5: CharacterCategory.CONTROL,  # '\x05'
    6: CharacterCategory.CONTROL,  # '\x06'
    7: CharacterCategory.CONTROL,  # '\x07'
    8: CharacterCategory.CONTROL,  # '\x08'
    9: CharacterCategory.CONTROL,  # '\t'
    10: CharacterCategory.LINE_BREAK,  # '\n'
    11: CharacterCategory.CONTROL,  # '\x0b'
    12: CharacterCategory.CONTROL,  # '\x0c'
    13: CharacterCategory.LINE_BREAK,  # '\r'
    14: CharacterCategory.CONTROL,  # '\x0e'
    15: CharacterCategory.CONTROL,  # '\x0f'
    16: CharacterCategory.CONTROL,  # '\x10'
    17: CharacterCategory.CONTROL,  # '\x11'
    18: CharacterCategory.CONTROL,  # '\x12'
    19: CharacterCategory.CONTROL,  # '\x13'
    20: CharacterCategory.CONTROL,  # '\x14'
    21: CharacterCategory.CONTROL,  # '\x15'
    22: CharacterCategory.CONTROL,  # '\x16'
    23: CharacterCategory.CONTROL,  # '\x17'
    24: CharacterCategory.CONTROL,  # '\x18'
    25: CharacterCategory.CONTROL,  # '\x19'
    26: CharacterCategory.CONTROL,  # '\x1a'
    27: CharacterCategory.CONTROL,  # '\x1b'
    28: CharacterCategory.CONTROL,  # '\x1c'
    29: CharacterCategory.CONTROL,  # '\x1d'
    30: CharacterCategory.CONTROL,  # '\x1e'
    31: CharacterCategory.CONTROL,  # '\x1f'
    32: CharacterCategory.CONTROL,  # ' '
    33: CharacterCategory.SYMBOL,  # '!'
    34: CharacterCategory.SYMBOL,  # '"'
    35: CharacterCategory.SYMBOL,  # '#'
    36: CharacterCategory.SYMBOL,  # '$'
    37: CharacterCategory.SYMBOL,  # '%'
    38: CharacterCategory.SYMBOL,  # '&'
    39: CharacterCategory.SYMBOL,  # "'"
    40: CharacterCategory.SYMBOL,  # '('
    41: CharacterCategory.SYMBOL,  # ')'
    42: CharacterCategory.SYMBOL,  # '*'
    43: CharacterCategory.SYMBOL,  # '+'
    44: CharacterCategory.SYMBOL,  # ','
    45: CharacterCategory.SYMBOL,  # '-'
    46: CharacterCategory.SYMBOL,  # '.'
    47: CharacterCategory.SYMBOL,  # '/'
    48: CharacterCategory.DIGIT,  # '0'
    49: CharacterCategory.DIGIT,  # '1'
    50: CharacterCategory.DIGIT,  # '2'
    51: CharacterCategory.DIGIT,  # '3'
    52: CharacterCategory.DIGIT,  # '4'
    53: CharacterCategory.DIGIT,  # '5'
    54: CharacterCategory.DIGIT,  # '6'
    55: CharacterCategory.DIGIT,  # '7'
    56: CharacterCategory.DIGIT,  # '8'
    57: CharacterCategory.DIGIT,  # '9'
    58: CharacterCategory.SYMBOL,  # ':'
    59: CharacterCategory.SYMBOL,  # ';'
    60: CharacterCategory.SYMBOL,  # '<'
    61: CharacterCategory.SYMBOL,  # '='
    62: CharacterCategory.SYMBOL,  # '>'
    63: CharacterCategory.SYMBOL,  # '?'
    64: CharacterCategory.SYMBOL,  # '@'
    65: 25,  # 'A'
    66: 28,  # 'B'
    67: 45,  # 'C'
    68: 26,  # 'D'
    69: 30,  # 'E'
    70: 33,  # 'F'
    71: 35,  # 'G'
    72: 38,  # 'H'
    73: 37,  # 'I'
    74: 51,  # 'J'
    75: 32,  # 'K'
    76: 41,  # 'L'
    77: 29,  # 'M'
    78: 43,  # 'N'
    79: 49,  # 'O'
    80: 34,  # 'P'
    81: 55,  # 'Q'
    82: 40,  # 'R'
    83: 22,  # 'S'
    84: 36,  # 'T'
    85: 46,  # 'U'
    86: 42,  # 'V'
    87: 31,  # 'W'
    88: 57,  # 'X'
    89: 56,  # 'Y'
    90: 47,  # 'Z'
    91: CharacterCategory.SYMBOL,  # '['
    92: CharacterCategory.SYMBOL,  # '\\'
    93: CharacterCategory.SYMBOL,  # ']'
    94: CharacterCategory.SYMBOL,  # '^'
    95: CharacterCategory.SYMBOL,  # '_'
    96: CharacterCategory.SYMBOL,  # '`'
    97: 7,  # 'a'
    98: 16,  # 'b'
    99: 13,  # 'c'
    100: 8,  # 'd'
    101: 1,  # 'e'
    102: 17,  # 'f'
    103: 14,  # 'g'
    104: 9,  # 'h'
    105: 3,  # 'i'
    106: 50,  # 'j'
    107: 18,  # 'k'
    108: 11,  # 'l'
    109: 15,  # 'm'
    110: 2,  # 'n'
    111: 12,  # 'o'
    112: 21,  # 'p'
    113: 53,  # 'q'
    114: 4,  # 'r'
    115: 6,  # 's'
    116: 5,  # 't'
    117: 10,  # 'u'
    118: 23,  # 'v'
    119: 19,  # 'w'
    120: 52,  # 'x'
    121: 44,  # 'y'
    122: 20,  # 'z'
    123: CharacterCategory.SYMBOL,  # '{'
    124: CharacterCategory.SYMBOL,  # '|'
    125: CharacterCategory.SYMBOL,  # '}'
    126: CharacterCategory.SYMBOL,  # '~'
    127: CharacterCategory.CONTROL,  # '\x7f'
    128: CharacterCategory.CONTROL,  # '\x80'
    129: CharacterCategory.CONTROL,  # '\x81'
    130: CharacterCategory.CONTROL,  # '\x82'
    131: CharacterCategory.CONTROL,  # '\x83'
    132: CharacterCategory.CONTROL,  # '\x84'
    133: CharacterCategory.CONTROL,  # '\x85'
    134: CharacterCategory.CONTROL,  # '\x86'
    135: CharacterCategory.CONTROL,  # '\x87'
    136: CharacterCategory.CONTROL,  # '\x88'
    137: CharacterCategory.CONTROL,  # '\x89'
    138: CharacterCategory.CONTROL,  # '\x8a'
    139: CharacterCategory.CONTROL,  # '\x8b'
    140: CharacterCategory.CONTROL,  # '\x8c'
    141: CharacterCategory.CONTROL,  # '\x8d'
    142: CharacterCategory.CONTROL,  # '\x8e'
    143: CharacterCategory.CONTROL,  # '\x8f'
    144: CharacterCategory.CONTROL,  # '\x90'
    145: CharacterCategory.CONTROL,  # '\x91'
    146: CharacterCategory.CONTROL,  # '\x92'
    147: CharacterCategory.CONTROL,  # '\x93'
    148: CharacterCategory.CONTROL,  # '\x94'
    149: CharacterCategory.CONTROL,  # '\x95'
    150: CharacterCategory.CONTROL,  # '\x96'
    151: CharacterCategory.CONTROL,  # '\x97'
    152: CharacterCategory.CONTROL,  # '\x98'
    153: CharacterCategory.CONTROL,  # '\x99'
    154: CharacterCategory.CONTROL,  # '\x9a'
    155: CharacterCategory.CONTROL,  # '\x9b'
    156: CharacterCategory.CONTROL,  # '\x9c'
    157: CharacterCategory.CONTROL,  # '\x9d'
    158: CharacterCategory.CONTROL,  # '\x9e'
    159: CharacterCategory.CONTROL,  # '\x9f'
    160: CharacterCategory.CONTROL,  # '\xa0'
    161: CharacterCategory.SYMBOL,  # '¡'
    162: CharacterCategory.SYMBOL,  # '¢'
    163: CharacterCategory.SYMBOL,  # '£'
    164: CharacterCategory.SYMBOL,  # '¤'
    165: CharacterCategory.SYMBOL,  # '¥'
    166: CharacterCategory.SYMBOL,  # '¦'
    167: CharacterCategory.SYMBOL,  # '§'
    168: CharacterCategory.SYMBOL,  # '¨'
    169: CharacterCategory.SYMBOL,  # '©'
    170: 0,  # 'ª'
    171: CharacterCategory.SYMBOL,  # '«'
    172: CharacterCategory.SYMBOL,  # '¬'
    173: CharacterCategory.CONTROL,  # '\xad'
    174: CharacterCategory.SYMBOL,  # '®'
    175: CharacterCategory.SYMBOL,  # '¯'
    176: CharacterCategory.SYMBOL,  # '°'
    177: CharacterCategory.SYMBOL,  # '±'
    178: CharacterCategory.DIGIT,  # '²'
    179: CharacterCategory.DIGIT,  # '³'
    180: CharacterCategory.SYMBOL,  # '´'
    181: 0,  # 'µ'
    182: CharacterCategory.SYMBOL,  # '¶'
    183: CharacterCategory.SYMBOL,  # '·'
    184: CharacterCategory.SYMBOL,  # '¸'
    185: CharacterCategory.DIGIT,  # '¹'
    186: 0,  # 'º'
    187: CharacterCategory.SYMBOL,  # '»'
    188: CharacterCategory.DIGIT,  # '¼'
    189: CharacterCategory.DIGIT,  # '½'
    190: CharacterCategory.DIGIT,  # '¾'
    191: CharacterCategory.SYMBOL,  # '¿'
    192: 0,  # 'À'
    193: 0,  # 'Á'
    194: 0,  # 'Â'
    195: 0,  # 'Ã'
    196: 59,  # 'Ä'
    197: 0,  # 'Å'
    198: 0,  # 'Æ'
    199: 0,  # 'Ç'
    200: 0,  # 'È'
    201: 0,  # 'É'
    202: 0,  # 'Ê'
    203: 0,  # 'Ë'
    204: 0,  # 'Ì'
    205: 0,  # 'Í'
    206: 0,  # 'Î'
    207: 0,  # 'Ï'
    208: 0,  # 'Ð'
    209: 0,  # 'Ñ'
    210: 0,  # 'Ò'
    211: 0,  # 'Ó'
    212: 0,  # 'Ô'
    213: 0,  # 'Õ'
    214: 58,  # 'Ö'
    215: CharacterCategory.SYMBOL,  # '×'
    216: 0,  # 'Ø'
    217: 0,  # 'Ù'
    218: 0,  # 'Ú'
    219: 0,  # 'Û'
    220: 54,  # 'Ü'
    221: 0,  # 'Ý'
    222: 0,  # 'Þ'
    223: 48,  # 'ß'
    224: 0,  # 'à'
    225: 0,  # 'á'
    226: 0,  # 'â'
    227: 0,  # 'ã'
    228: 27,  # 'ä'
    229: 0,  # 'å'
    230: 0,  # 'æ'
    231: 0,  # 'ç'
    232: 0,  # 'è'
    233: 0,  # 'é'
    234: 0,  # 'ê'
    235: 0,  # 'ë'
    236: 0,  # 'ì'
    237: 0,  # 'í'
    238: 0,  # 'î'
    239: 0,  # 'ï'
    240: 0,  # 'ð'
    241: 0,  # 'ñ'
    242: 0,  # 'ò'
    243: 0,  # 'ó'
    244: 0,  # 'ô'
    245: 0,  # 'õ'
    246: 39,  # 'ö'
    247: CharacterCategory.SYMBOL,  # '÷'
    248: 0,  # 'ø'
    249: 0,  # 'ù'
    250: 0,  # 'ú'
    251: 0,  # 'û'
    252: 24,  # 'ü'
    253: 0,  # 'ý'
    254: 0,  # 'þ'
    255: 0,  # 'ÿ'
}

ISO_8859_1_GERMAN_MODEL = SingleByteCharSetModel(
    charset_name="ISO-8859-1",
    language="German",
    char_to_order_map=ISO_8859_1_GERMAN_CHAR_TO_ORDER,
    language_model=GERMAN_LANG_MODEL,
    typical_positive_ratio=0.9706854667596968,
    keep_ascii_letters=True,
    alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyzÄÖÜßäöüẞ",
)

ISO_8859_15_GERMAN_CHAR_TO_ORDER = {
    0: CharacterCategory.CONTROL,  # '\x00'
    1: CharacterCategory.CONTROL,  # '\x01'
    2: CharacterCategory.CONTROL,  # '\x02'
    3: CharacterCategory.CONTROL,  # '\x03'
    4: CharacterCategory.CONTROL,  # '\x04'
    5: CharacterCategory.CONTROL,  # '\x05'
    6: CharacterCategory.CONTROL,  # '\x06'
    7: CharacterCategory.CONTROL,  # '\x07'
    8: CharacterCategory.CONTROL,  # '\x08'
    9: CharacterCategory.CONTROL,  # '\t'
    10: CharacterCategory.LINE_BREAK,  # '\n'
    11: CharacterCategory.CONTROL,  # '\x0b'
    12: CharacterCategory.CONTROL,  # '\x0c'
    13: CharacterCategory.LINE_BREAK,  # '\r'
    14: CharacterCategory.CONTROL,  # '\x0e'
    15: CharacterCategory.CONTROL,  # '\x0f'
    16: CharacterCategory.CONTROL,  # '\x10'
    17: CharacterCategory.CONTROL,  # '\x11'
    18: CharacterCategory.CONTROL,  # '\x12'
    19: CharacterCategory.CONTROL,  # '\x13'
    20: CharacterCategory.CONTROL,  # '\x14'
    21: CharacterCategory.CONTROL,  # '\x15'
    22: CharacterCategory.CONTROL,  # '\x16'
    23: CharacterCategory.CONTROL,  # '\x17'
    24: CharacterCategory.CONTROL,  # '\x18'
    25: CharacterCategory.CONTROL,  # '\x19'
    26: CharacterCategory.CONTROL,  # '\x1a'
    27: CharacterCategory.CONTROL,  # '\x1b'
    28: CharacterCategory.CONTROL,  # '\x1c'
    29: CharacterCategory.CONTROL,  # '\x1d'
    30: CharacterCategory.CONTROL,  # '\x1e'
    31: CharacterCategory.CONTROL,  # '\x1f'
    32: CharacterCategory.CONTROL,  # ' '
    33: CharacterCategory.SYMBOL,  # '!'
    34: CharacterCategory.SYMBOL,  # '"'
    35: CharacterCategory.SYMBOL,  # '#'
    36: CharacterCategory.SYMBOL,  # '$'
    37: CharacterCategory.SYMBOL,  # '%'
    38: CharacterCategory.SYMBOL,  # '&'
    39: CharacterCategory.SYMBOL,  # "'"
    40: CharacterCategory.SYMBOL,  # '('
    41: CharacterCategory.SYMBOL,  # ')'
    42: CharacterCategory.SYMBOL,  # '*'
    43: CharacterCategory.SYMBOL,  # '+'
    44: CharacterCategory.SYMBOL,  # ','
    45: CharacterCategory.SYMBOL,  # '-'
    46: CharacterCategory.SYMBOL,  # '.'
    47: CharacterCategory.SYMBOL,  # '/'
    48: CharacterCategory.DIGIT,  # '0'
    49: CharacterCategory.DIGIT,  # '1'
    50: CharacterCategory.DIGIT,  # '2'
    51: CharacterCategory.DIGIT,  # '3'
    52: CharacterCategory.DIGIT,  # '4'
    53: CharacterCategory.DIGIT,  # '5'
    54: CharacterCategory.DIGIT,  # '6'
    55: CharacterCategory.DIGIT,  # '7'
    56: CharacterCategory.DIGIT,  # '8'
    57: CharacterCategory.DIGIT,  # '9'
    58: CharacterCategory.SYMBOL,  # ':'
    59: CharacterCategory.SYMBOL,  # ';'
    60: CharacterCategory.SYMBOL,  # '<'
    61: CharacterCategory.SYMBOL,  # '='
    62: CharacterCategory.SYMBOL,  # '>'
    63: CharacterCategory.SYMBOL,  # '?'
    64: CharacterCategory.SYMBOL,  # '@'
    65: 25,  # 'A'
    66: 28,  # 'B'
    67: 45,  # 'C'
    68: 26,  # 'D'
    69: 30,  # 'E'
    70: 33,  # 'F'
    71: 35,  # 'G'
    72: 38,  # 'H'
    73: 37,  # 'I'
    74: 51,  # 'J'
    75: 32,  # 'K'
    76: 41,  # 'L'
    77: 29,  # 'M'
    78: 43,  # 'N'
    79: 49,  # 'O'
    80: 34,  # 'P'
    81: 55,  # 'Q'
    82: 40,  # 'R'
    83: 22,  # 'S'
    84: 36,  # 'T'
    85: 46,  # 'U'
    86: 42,  # 'V'
    87: 31,  # 'W'
    88: 57,  # 'X'
    89: 56,  # 'Y'
    90: 47,  # 'Z'
    91: CharacterCategory.SYMBOL,  # '['
    92: CharacterCategory.SYMBOL,  # '\\'
    93: CharacterCategory.SYMBOL,  # ']'
    94: CharacterCategory.SYMBOL,  # '^'
    95: CharacterCategory.SYMBOL,  # '_'
    96: CharacterCategory.SYMBOL,  # '`'
    97: 7,  # 'a'
    98: 16,  # 'b'
    99: 13,  # 'c'
    100: 8,  # 'd'
    101: 1,  # 'e'
    102: 17,  # 'f'
    103: 14,  # 'g'
    104: 9,  # 'h'
    105: 3,  # 'i'
    106: 50,  # 'j'
    107: 18,  # 'k'
    108: 11,  # 'l'
    109: 15,  # 'm'
    110: 2,  # 'n'
    111: 12,  # 'o'
    112: 21,  # 'p'
    113: 53,  # 'q'
    114: 4,  # 'r'
    115: 6,  # 's'
    116: 5,  # 't'
    117: 10,  # 'u'
    118: 23,  # 'v'
    119: 19,  # 'w'
    120: 52,  # 'x'
    121: 44,  # 'y'
    122: 20,  # 'z'
    123: CharacterCategory.SYMBOL,  # '{'
    124: CharacterCategory.SYMBOL,  # '|'
    125: CharacterCategory.SYMBOL,  # '}'
    126: CharacterCategory.SYMBOL,  # '~'
    127: CharacterCategory.CONTROL,  # '\x7f'
    128: CharacterCategory.CONTROL,  # '\x80'
    129: CharacterCategory.CONTROL,  # '\x81'
    130: CharacterCategory.CONTROL,  # '\x82'
    131: CharacterCategory.CONTROL,  # '\x83'
    132: CharacterCategory.CONTROL,  # '\x84'
    133: CharacterCategory.CONTROL,  # '\x85'
    134: CharacterCategory.CONTROL,  # '\x86'
    135: CharacterCategory.CONTROL,  # '\x87'
    136: CharacterCategory.CONTROL,  # '\x88'
    137: CharacterCategory.CONTROL,  # '\x89'
    138: CharacterCategory.CONTROL,  # '\x8a'
    139: CharacterCategory.CONTROL,  # '\x8b'
    140: CharacterCategory.CONTROL,  # '\x8c'
    141: CharacterCategory.CONTROL,  # '\x8d'
    142: CharacterCategory.CONTROL,  # '\x8e'
    143: CharacterCategory.CONTROL,  # '\x8f'
    144: CharacterCategory.CONTROL,  # '\x90'
    145: CharacterCategory.CONTROL,  # '\x91'
    146: CharacterCategory.CONTROL,  # '\x92'
    147: CharacterCategory.CONTROL,  # '\x93'
    148: CharacterCategory.CONTROL,  # '\x94'
    149: CharacterCategory.CONTROL,  # '\x95'
    150: CharacterCategory.CONTROL,  # '\x96'
    151: CharacterCategory.CONTROL,  # '\x97'
    152: CharacterCategory.CONTROL,  # '\x98'
    153: CharacterCategory.CONTROL,  # '\x99'
    154: CharacterCategory.CONTROL,  # '\x9a'
    155: CharacterCategory.CONTROL,  # '\x9b'
    156: CharacterCategory.CONTROL,  # '\x9c'
    157: CharacterCategory.CONTROL,  # '\x9d'
    158: CharacterCategory.CONTROL,  # '\x9e'
    159: CharacterCategory.CONTROL,  # '\x9f'
    160: CharacterCategory.CONTROL,  # '\xa0'
    161: CharacterCategory.SYMBOL,  # '¡'
    162: CharacterCategory.SYMBOL,  # '¢'
    163: CharacterCategory.SYMBOL,  # '£'
    164: CharacterCategory.SYMBOL,  # '€'
    165: CharacterCategory.SYMBOL,  # '¥'
    166: 0,  # 'Š'
    167: CharacterCategory.SYMBOL,  # '§'
    168: 0,  # 'š'
    169: CharacterCategory.SYMBOL,  # '©'
    170: 0,  # 'ª'
    171: CharacterCategory.SYMBOL,  # '«'
    172: CharacterCategory.SYMBOL,  # '¬'
    173: CharacterCategory.CONTROL,  # '\xad'
    174: CharacterCategory.SYMBOL,  # '®'
    175: CharacterCategory.SYMBOL,  # '¯'
    176: CharacterCategory.SYMBOL,  # '°'
    177: CharacterCategory.SYMBOL,  # '±'
    178: CharacterCategory.DIGIT,  # '²'
    179: CharacterCategory.DIGIT,  # '³'
    180: 0,  # 'Ž'
    181: 0,  # 'µ'
    182: CharacterCategory.SYMBOL,  # '¶'
    183: CharacterCategory.SYMBOL,  # '·'
    184: 0,  # 'ž'
    185: CharacterCategory.DIGIT,  # '¹'
    186: 0,  # 'º'
    187: CharacterCategory.SYMBOL,  # '»'
    188: 0,  # 'Œ'
    189: 0,  # 'œ'
    190: 0,  # 'Ÿ'
    191: CharacterCategory.SYMBOL,  # '¿'
    192: 0,  # 'À'
    193: 0,  # 'Á'
    194: 0,  # 'Â'
    195: 0,  # 'Ã'
    196: 59,  # 'Ä'
    197: 0,  # 'Å'
    198: 0,  # 'Æ'
    199: 0,  # 'Ç'
    200: 0,  # 'È'
    201: 0,  # 'É'
    202: 0,  # 'Ê'
    203: 0,  # 'Ë'
    204: 0,  # 'Ì'
    205: 0,  # 'Í'
    206: 0,  # 'Î'
    207: 0,  # 'Ï'
    208: 0,  # 'Ð'
    209: 0,  # 'Ñ'
    210: 0,  # 'Ò'
    211: 0,  # 'Ó'
    212: 0,  # 'Ô'
    213: 0,  # 'Õ'
    214: 58,  # 'Ö'
    215: CharacterCategory.SYMBOL,  # '×'
    216: 0,  # 'Ø'
    217: 0,  # 'Ù'
    218: 0,  # 'Ú'
    219: 0,  # 'Û'
    220: 54,  # 'Ü'
    221: 0,  # 'Ý'
    222: 0,  # 'Þ'
    223: 48,  # 'ß'
    224: 0,  # 'à'
    225: 0,  # 'á'
    226: 0,  # 'â'
    227: 0,  # 'ã'
    228: 27,  # 'ä'
    229: 0,  # 'å'
    230: 0,  # 'æ'
    231: 0,  # 'ç'
    232: 0,  # 'è'
    233: 0,  # 'é'
    234: 0,  # 'ê'
    235: 0,  # 'ë'
    236: 0,  # 'ì'
    237: 0,  # 'í'
    238: 0,  # 'î'
    239: 0,  # 'ï'
    240: 0,  # 'ð'
    241: 0,  # 'ñ'
    242: 0,  # 'ò'
    243: 0,  # 'ó'
    244: 0,  # 'ô'
    245: 0,  # 'õ'
    246: 39,  # 'ö'
    247: CharacterCategory.SYMBOL,  # '÷'
    248: 0,  # 'ø'
    249: 0,  # 'ù'
    250: 0,  # 'ú'
    251: 0,  # 'û'
    252: 24,  # 'ü'
    253: 0,  # 'ý'
    254: 0,  # 'þ'
    255: 0,  # 'ÿ'
}

ISO_8859_15_GERMAN_MODEL = SingleByteCharSetModel(
    charset_name="ISO-8859-15",
    language="German",
    char_to_order_map=ISO_8859_15_GERMAN_CHAR_TO_ORDER,
    language_model=GERMAN_LANG_MODEL,
    typical_positive_ratio=0.9706854667596968,
    keep_ascii_letters=True,
    alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyzÄÖÜßäöüẞ",
)

MACROMAN_GERMAN_CHAR_TO_ORDER = {
    0: CharacterCategory.CONTROL,  # '\x00'
    1: CharacterCategory.CONTROL,  # '\x01'
    2: CharacterCategory.CONTROL,  # '\x02'
    3: CharacterCategory.CONTROL,  # '\x03'
    4: CharacterCategory.CONTROL,  # '\x04'
    5: CharacterCategory.CONTROL,  # '\x05'
    6: CharacterCategory.CONTROL,  # '\x06'
    7: CharacterCategory.CONTROL,  # '\x07'
    8: CharacterCategory.CONTROL,  # '\x08'
    9: CharacterCategory.CONTROL,  # '\t'
    10: CharacterCategory.LINE_BREAK,  # '\n'
    11: CharacterCategory.CONTROL,  # '\x0b'
    12: CharacterCategory.CONTROL,  # '\x0c'
    13: CharacterCategory.LINE_BREAK,  # '\r'
    14: CharacterCategory.CONTROL,  # '\x0e'
    15: CharacterCategory.CONTROL,  # '\x0f'
    16: CharacterCategory.CONTROL,  # '\x10'
    17: CharacterCategory.CONTROL,  # '\x11'
    18: CharacterCategory.CONTROL,  # '\x12'
    19: CharacterCategory.CONTROL,  # '\x13'
    20: CharacterCategory.CONTROL,  # '\x14'
    21: CharacterCategory.CONTROL,  # '\x15'
    22: CharacterCategory.CONTROL,  # '\x16'
    23: CharacterCategory.CONTROL,  # '\x17'
    24: CharacterCategory.CONTROL,  # '\x18'
    25: CharacterCategory.CONTROL,  # '\x19'
    26: CharacterCategory.CONTROL,  # '\x1a'
    27: CharacterCategory.CONTROL,  # '\x1b'
    28: CharacterCategory.CONTROL,  # '\x1c'
    29: CharacterCategory.CONTROL,  # '\x1d'
    30: CharacterCategory.CONTROL,  # '\x1e'
    31: CharacterCategory.CONTROL,  # '\x1f'
    32: CharacterCategory.CONTROL,  # ' '
    33: CharacterCategory.SYMBOL,  # '!'
    34: CharacterCategory.SYMBOL,  # '"'
    35: CharacterCategory.SYMBOL,  # '#'
    36: CharacterCategory.SYMBOL,  # '$'
    37: CharacterCategory.SYMBOL,  # '%'
    38: CharacterCategory.SYMBOL,  # '&'
    39: CharacterCategory.SYMBOL,  # "'"
    40: CharacterCategory.SYMBOL,  # '('
    41: CharacterCategory.SYMBOL,  # ')'
    42: CharacterCategory.SYMBOL,  # '*'
    43: CharacterCategory.SYMBOL,  # '+'
    44: CharacterCategory.SYMBOL,  # ','
    45: CharacterCategory.SYMBOL,  # '-'
    46: CharacterCategory.SYMBOL,  # '.'
    47: CharacterCategory.SYMBOL,  # '/'
    48: CharacterCategory.DIGIT,  # '0'
    49: CharacterCategory.DIGIT,  # '1'
    50: CharacterCategory.DIGIT,  # '2'
    51: CharacterCategory.DIGIT,  # '3'
    52: CharacterCategory.DIGIT,  # '4'
    53: CharacterCategory.DIGIT,  # '5'
    54: CharacterCategory.DIGIT,  # '6'
    55: CharacterCategory.DIGIT,  # '7'
    56: CharacterCategory.DIGIT,  # '8'
    57: CharacterCategory.DIGIT,  # '9'
    58: CharacterCategory.SYMBOL,  # ':'
    59: CharacterCategory.SYMBOL,  # ';'
    60: CharacterCategory.SYMBOL,  # '<'
    61: CharacterCategory.SYMBOL,  # '='
    62: CharacterCategory.SYMBOL,  # '>'
    63: CharacterCategory.SYMBOL,  # '?'
    64: CharacterCategory.SYMBOL,  # '@'
    65: 25,  # 'A'
    66: 28,  # 'B'
    67: 45,  # 'C'
    68: 26,  # 'D'
    69: 30,  # 'E'
    70: 33,  # 'F'
    71: 35,  # 'G'
    72: 38,  # 'H'
    73: 37,  # 'I'
    74: 51,  # 'J'
    75: 32,  # 'K'
    76: 41,  # 'L'
    77: 29,  # 'M'
    78: 43,  # 'N'
    79: 49,  # 'O'
    80: 34,  # 'P'
    81: 55,  # 'Q'
    82: 40,  # 'R'
    83: 22,  # 'S'
    84: 36,  # 'T'
    85: 46,  # 'U'
    86: 42,  # 'V'
    87: 31,  # 'W'
    88: 57,  # 'X'
    89: 56,  # 'Y'
    90: 47,  # 'Z'
    91: CharacterCategory.SYMBOL,  # '['
    92: CharacterCategory.SYMBOL,  # '\\'
    93: CharacterCategory.SYMBOL,  # ']'
    94: CharacterCategory.SYMBOL,  # '^'
    95: CharacterCategory.SYMBOL,  # '_'
    96: CharacterCategory.SYMBOL,  # '`'
    97: 7,  # 'a'
    98: 16,  # 'b'
    99: 13,  # 'c'
    100: 8,  # 'd'
    101: 1,  # 'e'
    102: 17,  # 'f'
    103: 14,  # 'g'
    104: 9,  # 'h'
    105: 3,  # 'i'
    106: 50,  # 'j'
    107: 18,  # 'k'
    108: 11,  # 'l'
    109: 15,  # 'm'
    110: 2,  # 'n'
    111: 12,  # 'o'
    112: 21,  # 'p'
    113: 53,  # 'q'
    114: 4,  # 'r'
    115: 6,  # 's'
    116: 5,  # 't'
    117: 10,  # 'u'
    118: 23,  # 'v'
    119: 19,  # 'w'
    120: 52,  # 'x'
    121: 44,  # 'y'
    122: 20,  # 'z'
    123: CharacterCategory.SYMBOL,  # '{'
    124: CharacterCategory.SYMBOL,  # '|'
    125: CharacterCategory.SYMBOL,  # '}'
    126: CharacterCategory.SYMBOL,  # '~'
    127: CharacterCategory.CONTROL,  # '\x7f'
    128: 59,  # 'Ä'
    129: 0,  # 'Å'
    130: 0,  # 'Ç'
    131: 0,  # 'É'
    132: 0,  # 'Ñ'
    133: 58,  # 'Ö'
    134: 54,  # 'Ü'
    135: 0,  # 'á'
    136: 0,  # 'à'
    137: 0,  # 'â'
    138: 27,  # 'ä'
    139: 0,  # 'ã'
    140: 0,  # 'å'
    141: 0,  # 'ç'
    142: 0,  # 'é'
    143: 0,  # 'è'
    144: 0,  # 'ê'
    145: 0,  # 'ë'
    146: 0,  # 'í'
    147: 0,  # 'ì'
    148: 0,  # 'î'
    149: 0,  # 'ï'
    150: 0,  # 'ñ'
    151: 0,  # 'ó'
    152: 0,  # 'ò'
    153: 0,  # 'ô'
    154: 39,  # 'ö'
    155: 0,  # 'õ'
    156: 0,  # 'ú'
    157: 0,  # 'ù'
    158: 0,  # 'û'
    159: 24,  # 'ü'
    160: CharacterCategory.SYMBOL,  # '†'
    161: CharacterCategory.SYMBOL,  # '°'
    162: CharacterCategory.SYMBOL,  # '¢'
    163: CharacterCategory.SYMBOL,  # '£'
    164: CharacterCategory.SYMBOL,  # '§'
    165: CharacterCategory.SYMBOL,  # '•'
    166: CharacterCategory.SYMBOL,  # '¶'
    167: 48,  # 'ß'
    168: CharacterCategory.SYMBOL,  # '®'
    169: CharacterCategory.SYMBOL,  # '©'
    170: CharacterCategory.SYMBOL,  # '™'
    171: CharacterCategory.SYMBOL,  # '´'
    172: CharacterCategory.SYMBOL,  # '¨'
    173: CharacterCategory.SYMBOL,  # '≠'
    174: 0,  # 'Æ'
    175: 0,  # 'Ø'
    176: CharacterCategory.SYMBOL,  # '∞'
    177: CharacterCategory.SYMBOL,  # '±'
    178: CharacterCategory.SYMBOL,  # '≤'
    179: CharacterCategory.SYMBOL,  # '≥'
    180: CharacterCategory.SYMBOL,  # '¥'
    181: 0,  # 'µ'
    182: CharacterCategory.SYMBOL,  # '∂'
    183: CharacterCategory.SYMBOL,  # '∑'
    184: CharacterCategory.SYMBOL,  # '∏'
    185: 0,  # 'π'
    186: CharacterCategory.SYMBOL,  # '∫'
    187: 0,  # 'ª'
    188: 0,  # 'º'
    189: 0,  # 'Ω'
    190: 0,  # 'æ'
    191: 0,  # 'ø'
    192: CharacterCategory.SYMBOL,  # '¿'
    193: CharacterCategory.SYMBOL,  # '¡'
    194: CharacterCategory.SYMBOL,  # '¬'
    195: CharacterCategory.SYMBOL,  # '√'
    196: 0,  # 'ƒ'
    197: CharacterCategory.SYMBOL,  # '≈'
    198: CharacterCategory.SYMBOL,  # '∆'
    199: CharacterCategory.SYMBOL,  # '«'
    200: CharacterCategory.SYMBOL,  # '»'
    201: CharacterCategory.SYMBOL,  # '…'
    202: CharacterCategory.CONTROL,  # '\xa0'
    203: 0,  # 'À'
    204: 0,  # 'Ã'
    205: 0,  # 'Õ'
    206: 0,  # 'Œ'
    207: 0,  # 'œ'
    208: CharacterCategory.SYMBOL,  # '–'
    209: CharacterCategory.SYMBOL,  # '—'
    210: CharacterCategory.SYMBOL,  # '“'
    211: CharacterCategory.SYMBOL,  # '”'
    212: CharacterCategory.SYMBOL,  # '‘'
    213: CharacterCategory.SYMBOL,  # '’'
    214: CharacterCategory.SYMBOL,  # '÷'
    215: CharacterCategory.SYMBOL,  # '◊'
    216: 0,  # 'ÿ'
    217: 0,  # 'Ÿ'
    218: CharacterCategory.SYMBOL,  # '⁄'
    219: CharacterCategory.SYMBOL,  # '€'
    220: CharacterCategory.SYMBOL,  # '‹'
    221: CharacterCategory.SYMBOL,  # '›'
    222: 0,  # 'ﬁ'
    223: 0,  # 'ﬂ'
    224: CharacterCategory.SYMBOL,  # '‡'
    225: CharacterCategory.SYMBOL,  # '·'
    226: CharacterCategory.SYMBOL,  # '‚'
    227: CharacterCategory.SYMBOL,  # '„'
    228: CharacterCategory.SYMBOL,  # '‰'
    229: 0,  # 'Â'
    230: 0,  # 'Ê'
    231: 0,  # 'Á'
    232: 0,  # 'Ë'
    233: 0,  # 'È'
    234: 0,  # 'Í'
    235: 0,  # 'Î'
    236: 0,  # 'Ï'
    237: 0,  # 'Ì'
    238: 0,  # 'Ó'
    239: 0,  # 'Ô'
    240: CharacterCategory.CONTROL,  # '\uf8ff'
    241: 0,  # 'Ò'
    242: 0,  # 'Ú'
    243: 0,  # 'Û'
    244: 0,  # 'Ù'
    245: 0,  # 'ı'
    246: 0,  # 'ˆ'
    247: CharacterCategory.SYMBOL,  # '˜'
    248: CharacterCategory.SYMBOL,  # '¯'
    249: CharacterCategory.SYMBOL,  # '˘'
    250: CharacterCategory.SYMBOL,  # '˙'
    251: CharacterCategory.SYMBOL,  # '˚'
    252: CharacterCategory.SYMBOL,  # '¸'
    253: CharacterCategory.SYMBOL,  # '˝'
    254: CharacterCategory.SYMBOL,  # '˛'
    255: 0,  # 'ˇ'
}

MACROMAN_GERMAN_MODEL = SingleByteCharSetModel(
    charset_name="MacRoman",
    language="German",
    char_to_order_map=MACROMAN_GERMAN_CHAR_TO_ORDER,
    language_model=GERMAN_LANG_MODEL,
    typical_positive_ratio=0.9706854667596968,
    keep_ascii_letters=True,
    alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyzÄÖÜßäöüẞ",
)

WINDOWS_1252_GERMAN_CHAR_TO_ORDER = {
    0: CharacterCategory.CONTROL,  # '\x00'
    1: CharacterCategory.CONTROL,  # '\x01'
    2: CharacterCategory.CONTROL,  # '\x02'
    3: CharacterCategory.CONTROL,  # '\x03'
    4: CharacterCategory.CONTROL,  # '\x04'
    5: CharacterCategory.CONTROL,  # '\x05'
    6: CharacterCategory.CONTROL,  # '\x06'
    7: CharacterCategory.CONTROL,  # '\x07'
    8: CharacterCategory.CONTROL,  # '\x08'
    9: CharacterCategory.CONTROL,  # '\t'
    10: CharacterCategory.LINE_BREAK,  # '\n'
    11: CharacterCategory.CONTROL,  # '\x0b'
    12: CharacterCategory.CONTROL,  # '\x0c'
    13: CharacterCategory.LINE_BREAK,  # '\r'
    14: CharacterCategory.CONTROL,  # '\x0e'
    15: CharacterCategory.CONTROL,  # '\x0f'
    16: CharacterCategory.CONTROL,  # '\x10'
    17: CharacterCategory.CONTROL,  # '\x11'
    18: CharacterCategory.CONTROL,  # '\x12'
    19: CharacterCategory.CONTROL,  # '\x13'
    20: CharacterCategory.CONTROL,  # '\x14'
    21: CharacterCategory.CONTROL,  # '\x15'
    22: CharacterCategory.CONTROL,  # '\x16'
    23: CharacterCategory.CONTROL,  # '\x17'
    24: CharacterCategory.CONTROL,  # '\x18'
    25: CharacterCategory.CONTROL,  # '\x19'
    26: CharacterCategory.CONTROL,  # '\x1a'
    27: CharacterCategory.CONTROL,  # '\x1b'
    28: CharacterCategory.CONTROL,  # '\x1c'
    29: CharacterCategory.CONTROL,  # '\x1d'
    30: CharacterCategory.CONTROL,  # '\x1e'
    31: CharacterCategory.CONTROL,  # '\x1f'
    32: CharacterCategory.CONTROL,  # ' '
    33: CharacterCategory.SYMBOL,  # '!'
    34: CharacterCategory.SYMBOL,  # '"'
    35: CharacterCategory.SYMBOL,  # '#'
    36: CharacterCategory.SYMBOL,  # '$'
    37: CharacterCategory.SYMBOL,  # '%'
    38: CharacterCategory.SYMBOL,  # '&'
    39: CharacterCategory.SYMBOL,  # "'"
    40: CharacterCategory.SYMBOL,  # '('
    41: CharacterCategory.SYMBOL,  # ')'
    42: CharacterCategory.SYMBOL,  # '*'
    43: CharacterCategory.SYMBOL,  # '+'
    44: CharacterCategory.SYMBOL,  # ','
    45: CharacterCategory.SYMBOL,  # '-'
    46: CharacterCategory.SYMBOL,  # '.'
    47: CharacterCategory.SYMBOL,  # '/'
    48: CharacterCategory.DIGIT,  # '0'
    49: CharacterCategory.DIGIT,  # '1'
    50: CharacterCategory.DIGIT,  # '2'
    51: CharacterCategory.DIGIT,  # '3'
    52: CharacterCategory.DIGIT,  # '4'
    53: CharacterCategory.DIGIT,  # '5'
    54: CharacterCategory.DIGIT,  # '6'
    55: CharacterCategory.DIGIT,  # '7'
    56: CharacterCategory.DIGIT,  # '8'
    57: CharacterCategory.DIGIT,  # '9'
    58: CharacterCategory.SYMBOL,  # ':'
    59: CharacterCategory.SYMBOL,  # ';'
    60: CharacterCategory.SYMBOL,  # '<'
    61: CharacterCategory.SYMBOL,  # '='
    62: CharacterCategory.SYMBOL,  # '>'
    63: CharacterCategory.SYMBOL,  # '?'
    64: CharacterCategory.SYMBOL,  # '@'
    65: 25,  # 'A'
    66: 28,  # 'B'
    67: 45,  # 'C'
    68: 26,  # 'D'
    69: 30,  # 'E'
    70: 33,  # 'F'
    71: 35,  # 'G'
    72: 38,  # 'H'
    73: 37,  # 'I'
    74: 51,  # 'J'
    75: 32,  # 'K'
    76: 41,  # 'L'
    77: 29,  # 'M'
    78: 43,  # 'N'
    79: 49,  # 'O'
    80: 34,  # 'P'
    81: 55,  # 'Q'
    82: 40,  # 'R'
    83: 22,  # 'S'
    84: 36,  # 'T'
    85: 46,  # 'U'
    86: 42,  # 'V'
    87: 31,  # 'W'
    88: 57,  # 'X'
    89: 56,  # 'Y'
    90: 47,  # 'Z'
    91: CharacterCategory.SYMBOL,  # '['
    92: CharacterCategory.SYMBOL,  # '\\'
    93: CharacterCategory.SYMBOL,  # ']'
    94: CharacterCategory.SYMBOL,  # '^'
    95: CharacterCategory.SYMBOL,  # '_'
    96: CharacterCategory.SYMBOL,  # '`'
    97: 7,  # 'a'
    98: 16,  # 'b'
    99: 13,  # 'c'
    100: 8,  # 'd'
    101: 1,  # 'e'
    102: 17,  # 'f'
    103: 14,  # 'g'
    104: 9,  # 'h'
    105: 3,  # 'i'
    106: 50,  # 'j'
    107: 18,  # 'k'
    108: 11,  # 'l'
    109: 15,  # 'm'
    110: 2,  # 'n'
    111: 12,  # 'o'
    112: 21,  # 'p'
    113: 53,  # 'q'
    114: 4,  # 'r'
    115: 6,  # 's'
    116: 5,  # 't'
    117: 10,  # 'u'
    118: 23,  # 'v'
    119: 19,  # 'w'
    120: 52,  # 'x'
    121: 44,  # 'y'
    122: 20,  # 'z'
    123: CharacterCategory.SYMBOL,  # '{'
    124: CharacterCategory.SYMBOL,  # '|'
    125: CharacterCategory.SYMBOL,  # '}'
    126: CharacterCategory.SYMBOL,  # '~'
    127: CharacterCategory.CONTROL,  # '\x7f'
    128: CharacterCategory.SYMBOL,  # '€'
    129: CharacterCategory.UNDEFINED,  # None
    130: CharacterCategory.SYMBOL,  # '‚'
    131: 0,  # 'ƒ'
    132: CharacterCategory.SYMBOL,  # '„'
    133: CharacterCategory.SYMBOL,  # '…'
    134: CharacterCategory.SYMBOL,  # '†'
    135: CharacterCategory.SYMBOL,  # '‡'
    136: 0,  # 'ˆ'
    137: CharacterCategory.SYMBOL,  # '‰'
    138: 0,  # 'Š'
    139: CharacterCategory.SYMBOL,  # '‹'
    140: 0,  # 'Œ'
    141: CharacterCategory.UNDEFINED,  # None
    142: 0,  # 'Ž'
    143: CharacterCategory.UNDEFINED,  # None
    144: CharacterCategory.UNDEFINED,  # None
    145: CharacterCategory.SYMBOL,  # '‘'
    146: CharacterCategory.SYMBOL,  # '’'
    147: CharacterCategory.SYMBOL,  # '“'
    148: CharacterCategory.SYMBOL,  # '”'
    149: CharacterCategory.SYMBOL,  # '•'
    150: CharacterCategory.SYMBOL,  # '–'
    151: CharacterCategory.SYMBOL,  # '—'
    152: CharacterCategory.SYMBOL,  # '˜'
    153: CharacterCategory.SYMBOL,  # '™'
    154: 0,  # 'š'
    155: CharacterCategory.SYMBOL,  # '›'
    156: 0,  # 'œ'
    157: CharacterCategory.UNDEFINED,  # None
    158: 0,  # 'ž'
    159: 0,  # 'Ÿ'
    160: CharacterCategory.CONTROL,  # '\xa0'
    161: CharacterCategory.SYMBOL,  # '¡'
    162: CharacterCategory.SYMBOL,  # '¢'
    163: CharacterCategory.SYMBOL,  # '£'
    164: CharacterCategory.SYMBOL,  # '¤'
    165: CharacterCategory.SYMBOL,  # '¥'
    166: CharacterCategory.SYMBOL,  # '¦'
    167: CharacterCategory.SYMBOL,  # '§'
    168: CharacterCategory.SYMBOL,  # '¨'
    169: CharacterCategory.SYMBOL,  # '©'
    170: 0,  # 'ª'
    171: CharacterCategory.SYMBOL,  # '«'
    172: CharacterCategory.SYMBOL,  # '¬'
    173: CharacterCategory.CONTROL,  # '\xad'
    174: CharacterCategory.SYMBOL,  # '®'
    175: CharacterCategory.SYMBOL,  # '¯'
    176: CharacterCategory.SYMBOL,  # '°'
    177: CharacterCategory.SYMBOL,  # '±'
    178: CharacterCategory.DIGIT,  # '²'
    179: CharacterCategory.DIGIT,  # '³'
    180: CharacterCategory.SYMBOL,  # '´'
    181: 0,  # 'µ'
    182: CharacterCategory.SYMBOL,  # '¶'
    183: CharacterCategory.SYMBOL,  # '·'
    184: CharacterCategory.SYMBOL,  # '¸'
    185: CharacterCategory.DIGIT,  # '¹'
    186: 0,  # 'º'
    187: CharacterCategory.SYMBOL,  # '»'
    188: CharacterCategory.DIGIT,  # '¼'
    189: CharacterCategory.DIGIT,  # '½'
    190: CharacterCategory.DIGIT,  # '¾'
    191: CharacterCategory.SYMBOL,  # '¿'
    192: 0,  # 'À'
    193: 0,  # 'Á'
    194: 0,  # 'Â'
    195: 0,  # 'Ã'
    196: 59,  # 'Ä'
    197: 0,  # 'Å'
    198: 0,  # 'Æ'
    199: 0,  # 'Ç'
    200: 0,  # 'È'
    201: 0,  # 'É'
    202: 0,  # 'Ê'
    203: 0,  # 'Ë'
    204: 0,  # 'Ì'
    205: 0,  # 'Í'
    206: 0,  # 'Î'
    207: 0,  # 'Ï'
    208: 0,  # 'Ð'
    209: 0,  # 'Ñ'
    210: 0,  # 'Ò'
    211: 0,  # 'Ó'
    212: 0,  # 'Ô'
    213: 0,  # 'Õ'
    214: 58,  # 'Ö'
    215: CharacterCategory.SYMBOL,  # '×'
    216: 0,  # 'Ø'
    217: 0,  # 'Ù'
    218: 0,  # 'Ú'
    219: 0,  # 'Û'
    220: 54,  # 'Ü'
    221: 0,  # 'Ý'
    222: 0,  # 'Þ'
    223: 48,  # 'ß'
    224: 0,  # 'à'
    225: 0,  # 'á'
    226: 0,  # 'â'
    227: 0,  # 'ã'
    228: 27,  # 'ä'
    229: 0,  # 'å'
    230: 0,  # 'æ'
    231: 0,  # 'ç'
    232: 0,  # 'è'
    233: 0,  # 'é'
    234: 0,  # 'ê'
    235: 0,  # 'ë'
    236: 0,  # 'ì'
    237: 0,  # 'í'
    238: 0,  # 'î'
    239: 0,  # 'ï'
    240: 0,  # 'ð'
    241: 0,  # 'ñ'
    242: 0,  # 'ò'
    243: 0,  # 'ó'
    244: 0,  # 'ô'
    245: 0,  # 'õ'
    246: 39,  # 'ö'
    247: CharacterCategory.SYMBOL,  # '÷'
    248: 0,  # 'ø'
    249: 0,  # 'ù'
    250: 0,  # 'ú'
    251: 0,  # 'û'
    252: 24,  # 'ü'
    253: 0,  # 'ý'
    254: 0,  # 'þ'
    255: 0,  # 'ÿ'
}

WINDOWS_1252_GERMAN_MODEL = SingleByteCharSetModel(
    charset_name="WINDOWS-1252",
    language="German",
    char_to_order_map=WINDOWS_1252_GERMAN_CHAR_TO_ORDER,
    language_model=GERMAN_LANG_MODEL,
    typical_positive_ratio=0.9706854667596968,
    keep_ascii_letters=True,
    alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyzÄÖÜßäöüẞ",
)
